"""
Flask server for displaying supervisor output on web page.
"""

# built-in libraries
import html
import os
import secrets
from subprocess import Popen, PIPE, STDOUT, DEVNULL
import sys
import webbrowser

# third party imports
from flask import Flask, Response
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from flask_wtf.csrf import CSRFProtect

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import ssl_certificate
from thermostatsupervisor import flask_generic as flg
from thermostatsupervisor import supervise as sup
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util

# flask server
if env.is_windows_environment():
    # win server from Eclipse IDE:
    #     loopback will work to itself but not remote clients
    #     local IP works both itself and to remote Linux client.
    # win server from command line:
    #
    flask_ip_address = env.get_local_ip()
else:
    # Linux server from Thoney IDE: must update Thonny to run from root
    #   page opens on both loopback Linux and remote Win client, but
    #       no data loads.
    # flask_ip_address = '127.0.0.1'  # almost works from Linux client
    flask_ip_address = "0.0.0.0"
    # on Linux both methds are returning correct page header, but no data
FLASK_PORT = 5001  # note: ports below 1024 require root access on Linux
FLASK_USE_HTTPS = False  # HTTPS requires a cert to be installed.
if FLASK_USE_HTTPS:
    # Try to use generated SSL certificate, fallback to adhoc if needed
    FLASK_SSL_CERT = ssl_certificate.get_ssl_context(
        cert_file="supervisor_server.crt",
        key_file="supervisor_server.key",
        fallback_to_adhoc=True,
    )
    flask_kwargs = {"ssl_context": FLASK_SSL_CERT}
    FLASK_URL_PREFIX = "https://"
else:
    FLASK_SSL_CERT = None
    flask_kwargs = {}
    FLASK_URL_PREFIX = "http://"
flask_url = FLASK_URL_PREFIX + flask_ip_address + ":" + str(FLASK_PORT)

argv = []  # supervisor runtime args list


def create_app():
    """Create the flask object."""
    app_ = Flask(__name__)

    # Set a secret key for CSRF protection
    # In production, this should be set via environment variable
    secret_key = os.environ.get("SECRET_KEY")
    if not secret_key:
        # Generate a random secret key for development/testing

        secret_key = secrets.token_hex(32)
    app_.config["SECRET_KEY"] = secret_key

    # override JSONEncoder
    app_.json_encoder = flg.CustomJSONEncoder

    # api = Api(app)

    # api.add_resource(Controller, "/")
    return app_


# create the flask app
app = create_app()
# Initialize rate limiter
limiter = Limiter(
    get_remote_address,
    app=app,
    default_limits=["200 per day", "60 per hour"],
    storage_uri=env.get_flask_limiter_storage_uri(),
)
csrf = CSRFProtect(app)  # enable CSRF protection
ip_ban = flg.initialize_ipban(app)  # hacker BlockListing agent
flg.set_flask_cookie_config(app)
flg.print_flask_config(app)


@app.route("/favicon.ico")
def favicon():
    """Faviocon displayed in browser tab."""
    return app.send_static_file("honeywell.ico")


@app.route("/data")
@limiter.limit("1 per minute")
def index():
    """index route"""

    def run_supervise():
        sup.argv = argv  # pass runtime overrides to supervise
        api.uip = api.UserInputs(argv)
        thermostat_type = api.uip.get_user_inputs(
            api.uip.zone_name, api.input_flds.thermostat_type
        )
        zone = api.uip.get_user_inputs(api.uip.zone_name, api.input_flds.zone)
        measurement_cnt = api.uip.get_user_inputs(
            api.uip.zone_name, api.input_flds.measurements
        )
        title = (
            f"{thermostat_type} thermostat zone {zone}, "
            f"{measurement_cnt} measurements"
        )
        yield f"<!doctype html><title>{title}</title>"

        # runtime variabless
        executable = "python"
        dont_buffer = "-u"  # option to not buffer results
        run_module = "-m"  # option to reference package
        script = "thermostatsupervisor.supervise"
        if argv:
            # argv list override for unit testing
            arg_list = [executable, dont_buffer, run_module, script] + argv[1:]
        elif len(sys.argv) > 1:
            arg_list = [executable, dont_buffer, run_module, script] + sys.argv[1:]
        else:
            arg_list = [executable, dont_buffer, run_module, script]
        with Popen(
            arg_list,
            stdin=DEVNULL,
            stdout=PIPE,
            stderr=STDOUT,
            bufsize=1,
            universal_newlines=True,
            shell=True,
        ) as p_out:
            for i, line in enumerate(p_out.stdout):
                print(f"DEBUG: line {i}: {line}", file=sys.stderr)
                yield "<code>{}</code>".format(html.escape(line.rstrip("\n")))
                yield "<br>\n"

    return Response(run_supervise(), mimetype="text/html")


if __name__ == "__main__":
    # enable logging to STDERR for Flask
    util.log_stdout_to_stderr = True

    # show the page in browser
    webbrowser.open(flask_url, new=2)
    flg.schedule_ipban_block_list_report(ip_ban, debug_mode=False)
    app.run(
        host=flask_ip_address,
        port=FLASK_PORT,
        debug=False,  # True causes 2 tabs to open, enables auto-reload
        threaded=True,  # threaded=True may speed up rendering on web page
        ssl_context=FLASK_SSL_CERT,
    )



"""
Unit test for SHT31 flask path selection logic.

This test validates the fix for issue where SHT31 ThermostatClass
was not using the correct path (/unit) during unit tests.
"""

import unittest
from unittest.mock import patch

from thermostatsupervisor import sht31
from thermostatsupervisor import sht31_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class TestSHT31PathSelection(utc.UnitTest):
    """Test SHT31 flask path selection logic."""

    def setUp(self):
        """Set up test fixtures."""
        super().setUp()
        self.original_unit_test_mode = util.unit_test_mode

    def tearDown(self):
        """Clean up after tests."""
        super().tearDown()
        util.unit_test_mode = self.original_unit_test_mode

    @patch.object(sht31.ThermostatClass, "get_target_zone_id")
    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    def test_unit_test_zone_always_uses_unit_path(self, mock_spawn, mock_get_zone_id):
        """Test that zone 99 always uses unit test path."""
        mock_get_zone_id.return_value = "127.0.0.1"
        mock_spawn.return_value = None

        for unit_test_mode_setting in [False, True]:
            with self.subTest(unit_test_mode=unit_test_mode_setting):
                util.unit_test_mode = unit_test_mode_setting

                tstat = sht31.ThermostatClass(
                    sht31_config.UNIT_TEST_ZONE,
                    sht31_config.flask_folder.production,
                    verbose=False,
                )

                self.assertEqual(tstat.path, sht31_config.flask_folder.unit_test)
                self.assertIn("/unit", tstat.url)
                self.assertIn("seed=", tstat.url)

    @patch.object(sht31.ThermostatClass, "get_target_zone_id")
    def test_regular_zone_normal_mode_uses_production_path(self, mock_get_zone_id):
        """Test that regular zones use production path in normal mode."""
        mock_get_zone_id.return_value = "127.0.0.1"
        util.unit_test_mode = False

        tstat = sht31.ThermostatClass(
            1, sht31_config.flask_folder.production, verbose=False
        )

        self.assertEqual(tstat.path, sht31_config.flask_folder.production)
        self.assertIn("/data", tstat.url)
        self.assertNotIn("seed=", tstat.url)

    @patch.object(sht31.ThermostatClass, "get_target_zone_id")
    def test_regular_zone_unit_test_mode_uses_unit_path(self, mock_get_zone_id):
        """Test that regular zones use unit test path in unit test mode."""
        mock_get_zone_id.return_value = "127.0.0.1"
        util.unit_test_mode = True

        # This is the main fix being tested
        tstat = sht31.ThermostatClass(
            1, sht31_config.flask_folder.production, verbose=False
        )

        self.assertEqual(tstat.path, sht31_config.flask_folder.unit_test)
        self.assertIn("/unit", tstat.url)
        self.assertIn("seed=", tstat.url)

    @patch.object(sht31.ThermostatClass, "get_target_zone_id")
    def test_custom_path_preserved(self, mock_get_zone_id):
        """Test that custom paths are preserved."""
        mock_get_zone_id.return_value = "127.0.0.1"

        for unit_test_mode_setting in [False, True]:
            with self.subTest(unit_test_mode=unit_test_mode_setting):
                util.unit_test_mode = unit_test_mode_setting

                custom_path = "/custom"
                tstat = sht31.ThermostatClass(1, custom_path, verbose=False)

                # Custom paths should not be changed
                self.assertEqual(tstat.path, custom_path)
                self.assertIn("/custom", tstat.url)
                self.assertNotIn("seed=", tstat.url)

    @patch.object(sht31.ThermostatClass, "get_target_zone_id")
    def test_integration_test_scenario_fix(self, mock_get_zone_id):
        """
        Test the specific scenario that was failing.

        This reproduces the exact error scenario:
        - Zone 1 (from integration test argv)
        - Unit test mode enabled (as in test environment)
        - Should use /unit path, not /data
        """
        mock_get_zone_id.return_value = "bsl-pi0.ddns.net"
        util.unit_test_mode = True

        tstat = sht31.ThermostatClass(1, verbose=False)

        # Verify the URL contains the expected components
        self.assertIn("bsl-pi0.ddns.net:5000", tstat.url)
        self.assertIn("/unit", tstat.url)
        self.assertIn("measurements=10", tstat.url)
        self.assertIn("seed=127", tstat.url)

        # Most importantly, verify it does NOT contain /data
        self.assertNotIn("/data", tstat.url)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
Common Thermostat Class
"""

# built-ins
import datetime
import operator
import pprint
import statistics
import time
import traceback
from typing import Union

# local imports
from thermostatsupervisor import email_notification as eml
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util
from thermostatsupervisor import weather


DEGREE_SIGN = "\N{DEGREE SIGN}"

connection_ok = True  # global flag for connection OK.
server_spamming_detected = False  # global flag for pyhtcc server spamming


def reset_server_spamming_flag():
    """Reset the server spamming detection flag."""
    global server_spamming_detected
    server_spamming_detected = False


class ThermostatCommon:
    """Class methods common to all thermostat objects."""

    def __init__(self, *_, **__):
        self.verbose = False
        self.thermostat_type = "unknown"  # placeholder
        self.zone_number = util.BOGUS_INT  # placeholder
        self.device_id = util.BOGUS_INT  # placeholder
        self.ip_address = None  # placeholder
        global connection_ok  # noqa W603
        connection_ok = True

    def get_all_metadata(self, zone, retry=True):
        """
        Get all the current thermostat metadata.

        inputs:
            zone(int): target zone
            retry(bool): if True will retry once.
        returns:
            (dict) results dict.
        """
        raise NotImplementedError(
            "get_all_metadata is not implemented for this thermostat type"
        )

    def get_metadata(self, zone=None, trait=None, parameter=None):
        """Get thermostat meta data for zone.

        inputs:
            zone(str or int): specified zone
            trait(str): trait or parent key, if None will assume a non-nested
                        dict.
            parameter(str): target parameter, if None will return all.
        returns:
            (dict): dictionary of meta data.
        """
        raise NotImplementedError(
            "get_metadata is not implemented for this thermostat type"
        )

    def print_all_thermostat_metadata(self, zone):  # noqa R0201
        """
        Print initial meta data queried from thermostat for specified zone.

        inputs:
            zone(int): zone number
        returns:
            None
        """
        util.log_msg(
            f"WARNING: print_all_thermostat_metatdata({zone}) not yet "
            f"implemented for this thermostat type\n",
            mode=util.BOTH_LOG,
            func_name=1,
        )

    def exec_print_all_thermostat_metadata(self, func, args):
        """
        Print all metadata to screen.

        inputs:
            func(obj): function get metadata.
            args(list): argument list
        returns:
            (dict): return data
        """
        # dump metadata in a readable format
        return_data = func(*args)
        pprint_obj = pprint.PrettyPrinter(indent=4)
        print("\n")
        util.log_msg("raw thermostat meta data:", mode=util.BOTH_LOG, func_name=1)
        pprint_obj.pprint(return_data)
        return return_data


class ThermostatCommonZone:
    """Class methods common to all thermostat zones."""

    # supported thermostat modes and label text
    OFF_MODE = "OFF_MODE"
    HEAT_MODE = "HEAT_MODE"
    COOL_MODE = "COOL_MODE"
    AUTO_MODE = "AUTO_MODE"
    DRY_MODE = "DRY_MODE"
    FAN_MODE = "FAN_MODE"
    ECO_MODE = "MANUAL_ECO"
    UNKNOWN_MODE = "UNKNOWN_MODE"  # bypass set mode or unable to detect

    # modes where heat is applied
    heat_modes = [HEAT_MODE, AUTO_MODE, ECO_MODE]

    # modes where cooling is applied
    cool_modes = [COOL_MODE, DRY_MODE, AUTO_MODE, ECO_MODE]

    # modes in which setpoints apply
    controlled_modes = [HEAT_MODE, AUTO_MODE, COOL_MODE]

    system_switch_position = {
        # placeholder, will be tstat-specific
        UNKNOWN_MODE: util.BOGUS_INT,
        HEAT_MODE: util.BOGUS_INT - 1,
        COOL_MODE: util.BOGUS_INT - 2,
        AUTO_MODE: util.BOGUS_INT - 3,
        DRY_MODE: util.BOGUS_INT - 4,
        FAN_MODE: util.BOGUS_INT - 5,
        OFF_MODE: util.BOGUS_INT - 6,
        ECO_MODE: util.BOGUS_INT - 7,
    }
    max_scheduled_heat_allowed = 74  # warn if scheduled heat value exceeds.
    min_scheduled_cool_allowed = 68  # warn if scheduled cool value exceeds.
    tolerance_degrees_default = 2  # allowed override vs. the scheduled value.

    def __init__(self, *_, **__):
        self.verbose = False
        self.thermostat_type = "unknown"  # placeholder
        self.zone_number = util.BOGUS_INT  # placeholder
        self.zone_name = None  # placeholder
        self.device_id = util.BOGUS_INT  # placeholder
        self.poll_time_sec = util.BOGUS_INT  # placeholder
        self.session_start_time_sec = util.BOGUS_INT  # placeholder
        self.connection_time_sec = util.BOGUS_INT  # placeholder
        self.target_mode = "OFF_MODE"  # placeholder
        self.flag_all_deviations = False  #
        self.revert_deviations = True  # Revert deviation
        self.revert_all_deviations = False  # Revert only energy-wasting events
        self.temperature_is_deviated = False  # temp deviated from schedule
        self.display_temp = None  # current temperature in °F
        self.display_humidity = None  # current humidity in %RH
        self.humidity_is_available = False  # humidity supported flag
        self.hold_mode = False  # True = not following schedule
        self.hold_temporary = False
        self.zone_info = {}  # dict containing zone data

        # server data cache expiration parameters
        self.fetch_interval_sec = 10  # age of server data before refresh
        self.last_fetch_time = time.time() - 2 * self.fetch_interval_sec

        # abstraction vars and funcs, defined in query_thermostat_zone
        self.current_mode = None  # str representing mode
        self.current_setpoint = float(util.BOGUS_INT)  # current setpoint
        self.schedule_setpoint = float(util.BOGUS_INT)  # current scheduled setpoint
        self.tolerance_sign = 1  # +1 for heat, -1 for cool
        self.operator = operator.ne  # operator for deviation check
        self.tolerance_degrees = self.tolerance_degrees_default
        self.global_limit = util.BOGUS_INT  # global temp limit
        self.global_operator = operator.ne  # oper for global temp deviation
        self.revert_setpoint_func = self.function_not_supported
        self.get_setpoint_func = self.function_not_supported

    def query_thermostat_zone(self):
        """Return the current mode and set mode-specific parameters."""
        self._set_current_temperature_and_humidity()
        self._configure_mode_specific_parameters()
        self.temperature_is_deviated = self.is_temp_deviated_from_schedule()

    def _set_current_temperature_and_humidity(self):
        """Set current temperature and humidity values."""
        try:
            self.display_temp = self.validate_numeric(
                self.get_display_temp(), "get_display_temp"
            )
        except TypeError:
            util.log_msg(traceback.format_exc(), mode=util.BOTH_LOG, func_name=1)
            self.display_temp = None

        self.display_humidity = self.get_display_humidity()
        self.humidity_is_available = self.get_is_humidity_supported()

    def _configure_mode_specific_parameters(self):
        """Configure mode-specific parameters using strategy pattern."""
        mode_handlers = {
            self.is_heat_mode: self._configure_heat_mode,
            self.is_cool_mode: self._configure_cool_mode,
            self.is_dry_mode: self._configure_dry_mode,
            self.is_auto_mode: self._configure_auto_mode,
            self.is_eco_mode: self._configure_eco_mode,
            self.is_fan_mode: self._configure_fan_mode,
            self.is_off_mode: self._configure_off_mode,
        }

        for mode_checker, mode_configurator in mode_handlers.items():
            if mode_checker():
                mode_configurator()
                return

        # If no mode matches, raise error
        print(f"DEBUG: zone info: {self.zone_info}")
        raise ValueError("unknown thermostat mode")

    def _configure_heat_mode(self):
        """Configure parameters for heat mode."""
        self.current_mode = self.HEAT_MODE
        self.current_setpoint = float(self.get_heat_setpoint_raw())
        self.schedule_setpoint = int(self.get_schedule_heat_sp())
        self.tolerance_sign = 1
        self._set_deviation_behavior(operator.gt, operator.ne)
        self.global_limit = self.max_scheduled_heat_allowed
        self.global_operator = operator.gt
        self.revert_setpoint_func = self.set_heat_setpoint
        self.get_setpoint_func = self.get_heat_setpoint_raw

    def _configure_cool_mode(self):
        """Configure parameters for cool mode."""
        self.current_mode = self.COOL_MODE
        self.current_setpoint = float(self.get_cool_setpoint_raw())
        self.schedule_setpoint = float(self.get_schedule_cool_sp())
        self.tolerance_sign = -1
        self._set_deviation_behavior(operator.lt, operator.ne)
        self.global_limit = self.min_scheduled_cool_allowed
        self.global_operator = operator.lt
        self.revert_setpoint_func = self.set_cool_setpoint
        self.get_setpoint_func = self.get_cool_setpoint_raw

    def _configure_dry_mode(self):
        """Configure parameters for dry mode."""
        self.current_mode = self.DRY_MODE
        self.current_setpoint = float(self.get_cool_setpoint_raw())
        self.schedule_setpoint = float(self.get_schedule_cool_sp())
        self.tolerance_sign = -1
        self._set_deviation_behavior(operator.lt, operator.ne)
        self.global_limit = self.min_scheduled_cool_allowed
        self.global_operator = operator.lt
        self.revert_setpoint_func = self.function_not_supported
        self.get_setpoint_func = self.function_not_supported

    def _configure_auto_mode(self):
        """Configure parameters for auto mode."""
        self._configure_unsupported_mode(self.AUTO_MODE)

    def _configure_eco_mode(self):
        """Configure parameters for eco mode."""
        self._configure_unsupported_mode(self.ECO_MODE)

    def _configure_fan_mode(self):
        """Configure parameters for fan mode."""
        self._configure_unsupported_mode(self.FAN_MODE)

    def _configure_off_mode(self):
        """Configure parameters for off mode."""
        self._configure_unsupported_mode(self.OFF_MODE)

    def _configure_unsupported_mode(self, mode):
        """Configure parameters for modes that don't support setpoint control."""
        self.current_mode = mode
        self.current_setpoint = util.BOGUS_INT
        self.schedule_setpoint = util.BOGUS_INT
        self.tolerance_sign = 1
        self.operator = operator.ne
        self.global_limit = util.BOGUS_INT
        self.global_operator = operator.ne
        self.revert_setpoint_func = self.function_not_supported
        self.get_setpoint_func = self.function_not_supported

    def _set_deviation_behavior(self, normal_operator, deviation_operator):
        """Set operator and tolerance based on deviation flag."""
        if self.flag_all_deviations:
            self.operator = deviation_operator
            self.tolerance_degrees = 0  # disable tolerance
        else:
            self.operator = normal_operator

    def is_temp_deviated_from_schedule(self):
        """
        Return True if temperature is deviated from current schedule.

        inputs:
            None:
        returns:
            (bool): True if temp is deviated from schedule.
        """
        return self.operator(
            self.current_setpoint,
            self.schedule_setpoint + self.tolerance_sign * self.tolerance_degrees,
        )

    def get_current_mode(
        self, session_count, poll_count, print_status=True, flag_all_deviations=False
    ):
        """
        Determine whether thermostat is following schedule or if it has been
        deviated from schedule.

        inputs:
            session_count(int): session number (connection #) for reporting
            poll_count(int): poll number for reporting
            print_status(bool):  True to print status line
            flag_all_deviations(bool):  True: flag all deviations
                                        False(default): only flag energy
                                                        consuming deviations,
                                                        e.g. heat setpoint
                                                        above schedule,
                                                        cool setpoint
                                                        below schedule
        returns:
            dictionary of heat/cool mode status, deviation status,
            and hold status
        """

        return_buffer = {
            "heat_mode": util.BOGUS_BOOL,  # in heating mode
            "cool_mode": util.BOGUS_BOOL,  # in cooling mode
            "heat_deviation": util.BOGUS_BOOL,  # True: heat is deviated above
            "cool_deviation": util.BOGUS_BOOL,  # True: cool is deviated below
            "hold_mode": util.BOGUS_BOOL,  # True if hold is enabled
            "status_msg": "",  # status message
        }

        self.flag_all_deviations = flag_all_deviations
        self.query_thermostat_zone()

        # warning email if set point is outside global limit
        self.warn_if_outside_global_limit(
            self.current_setpoint,
            self.global_limit,
            self.global_operator,
            self.current_mode,
        )

        if self.is_temp_deviated_from_schedule() and self.is_controlled_mode():
            status_msg = (
                f"[{self.current_mode.upper()} deviation] act temp="
                f"{util.temp_value_with_units(self.display_temp)}"
            )
        else:
            status_msg = (
                f"[following schedule] act temp="
                f"{util.temp_value_with_units(self.display_temp)}"
            )

        # add humidity if available
        if self.humidity_is_available:
            status_msg += (
                f", act humidity="
                f"{util.humidity_value_with_units(self.display_humidity)}"
            )

        # add hold information
        if self.is_temp_deviated_from_schedule() and self.is_controlled_mode():
            self.hold_mode = True  # True = not following schedule
            self.hold_temporary = self.get_temporary_hold_until_time() > 0
            status_msg += f" ({['persistent', 'temporary'][self.hold_temporary]})"
        else:
            self.hold_mode = False
            self.hold_temporary = False

        # add setpoints if in heat or cool mode
        if self.is_heat_mode() or self.is_cool_mode():
            status_msg += (
                f", set point="
                f"{util.temp_value_with_units(self.schedule_setpoint)}, "
                f"tolerance="
                f"{util.temp_value_with_units(self.tolerance_degrees, precision=0)}, "
                f"override="
                f"{util.temp_value_with_units(self.current_setpoint)}"
            )

        date_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        full_status_msg = (
            f"{date_str}: "
            f"(tstat:{self.thermostat_type}, zone:{self.zone_name}, "
            f"session:{session_count}, poll:{poll_count}) "
            f"{self.current_mode.upper()} {status_msg}"
        )
        if print_status:
            util.log_msg(full_status_msg, mode=util.BOTH_LOG)

        self.store_current_mode()

        # return status
        return_buffer["heat_mode"] = self.is_heat_mode()
        return_buffer["cool_mode"] = self.is_cool_mode()
        return_buffer["heat_deviation"] = self.is_heat_deviation()
        return_buffer["cool_deviation"] = self.is_cool_deviation()
        return_buffer["hold_mode"] = self.hold_mode
        return_buffer["status_msg"] = full_status_msg
        return return_buffer

    def set_mode(self, target_mode):
        """
        Set the thermostat mode and apply the scheduled setpoint.

        This method sets the thermostat to the specified mode and applies
        the appropriate scheduled setpoint for that mode.

        inputs:
            target_mode(str): target mode, one of the supported mode constants
                             (HEAT_MODE, COOL_MODE, AUTO_MODE, etc.)
        returns:
            True if successful, False otherwise
        """
        if self.verbose:
            util.log_msg(
                f"Setting thermostat mode to {target_mode}",
                mode=util.BOTH_LOG,
                func_name=1,
            )

        if not self._is_valid_mode(target_mode):
            return False

        try:
            self._apply_mode_setpoint(target_mode)
            if self.verbose:
                util.log_msg(
                    f"Mode set operation completed for {target_mode}",
                    mode=util.BOTH_LOG,
                    func_name=1,
                )
            return True

        except Exception as e:
            util.log_msg(
                f"ERROR setting mode to {target_mode}: {e}",
                mode=util.BOTH_LOG,
                func_name=1,
            )
            return False

    def _is_valid_mode(self, target_mode):
        """Validate if the target mode is supported."""
        valid_modes = [
            self.HEAT_MODE,
            self.COOL_MODE,
            self.AUTO_MODE,
            self.DRY_MODE,
            self.FAN_MODE,
            self.OFF_MODE,
            self.ECO_MODE,
            self.UNKNOWN_MODE,
        ]

        if target_mode not in valid_modes:
            util.log_msg(
                f"ERROR: Invalid target mode '{target_mode}'. "
                f"Valid modes: {valid_modes}",
                mode=util.BOTH_LOG,
                func_name=1,
            )
            return False
        return True

    def _apply_mode_setpoint(self, target_mode):
        """Apply scheduled setpoint based on target mode."""
        mode_handlers = {
            self.HEAT_MODE: self._apply_heat_setpoint,
            self.COOL_MODE: self._apply_cool_setpoint,
            self.UNKNOWN_MODE: self._apply_unknown_mode,
        }

        if target_mode in mode_handlers:
            mode_handlers[target_mode]()
        elif target_mode in [
            self.AUTO_MODE,
            self.DRY_MODE,
            self.FAN_MODE,
            self.OFF_MODE,
            self.ECO_MODE,
        ]:
            self._apply_no_setpoint_mode(target_mode)

    def _apply_heat_setpoint(self):
        """Apply scheduled heat setpoint."""
        scheduled_setpoint = self.get_schedule_heat_sp()
        if scheduled_setpoint != util.BOGUS_INT:
            self.set_heat_setpoint(int(scheduled_setpoint))
            if self.verbose:
                util.log_msg(
                    f"Applied scheduled heat setpoint: "
                    f"{util.temp_value_with_units(scheduled_setpoint)}",
                    mode=util.BOTH_LOG,
                    func_name=1,
                )

    def _apply_cool_setpoint(self):
        """Apply scheduled cool setpoint."""
        scheduled_setpoint = self.get_schedule_cool_sp()
        if scheduled_setpoint != util.BOGUS_INT:
            self.set_cool_setpoint(int(scheduled_setpoint))
            if self.verbose:
                util.log_msg(
                    f"Applied scheduled cool setpoint: "
                    f"{util.temp_value_with_units(scheduled_setpoint)}",
                    mode=util.BOTH_LOG,
                    func_name=1,
                )

    def _apply_no_setpoint_mode(self, target_mode):
        """Handle modes that don't require setpoint changes."""
        if self.verbose:
            util.log_msg(
                f"Mode {target_mode} set without setpoint changes",
                mode=util.BOTH_LOG,
                func_name=1,
            )

    def _apply_unknown_mode(self):
        """Handle unknown mode (bypass mode)."""
        if self.verbose:
            util.log_msg(
                f"Mode {self.UNKNOWN_MODE} - no action taken (bypass mode)",
                mode=util.BOTH_LOG,
                func_name=1,
            )

    def store_current_mode(self):
        """Save the current mode to cache."""
        if self.is_heat_mode():
            self.current_mode = self.HEAT_MODE
        elif self.is_cool_mode():
            self.current_mode = self.COOL_MODE
        elif self.is_dry_mode():
            self.current_mode = self.DRY_MODE
        elif self.is_auto_mode():
            self.current_mode = self.AUTO_MODE
        elif self.is_eco_mode():
            self.current_mode = self.ECO_MODE
        elif self.is_fan_mode():
            self.current_mode = self.FAN_MODE
        elif self.is_off_mode():
            self.current_mode = self.OFF_MODE
        else:
            raise ValueError("unknown thermostat mode")

    def validate_numeric(self, input_val, parameter_name):
        """
        Validate value returned is numeric, otherwise raise exception.

        inputs:
            input_val: input value of unknown type.
            parameter_name(str): parameter name.
        returns:
            (int, float): pass thru value if numeric, else raise exception.
        """
        if not isinstance(input_val, (int, float)):
            raise TypeError(
                f"value returned for parameter '{parameter_name}' is "
                f"type {type(input_val)}, expected int or float"
            )
        return input_val

    def warn_if_outside_global_limit(self, setpoint, limit_value, oper, label) -> bool:
        """
        Send warning email if setpoint is outside of global limits.

        inputs:
            setpoint(int): setpoint value.
            limit_value(int): the limit value
            oper(operator):  the operator, either operator.gt or operator.lt
            label(str): label for warning message denoting the mode
        returns:
            (bool): result of check
        """
        if oper == operator.gt:  # pylint: disable=W0143
            level = "above max"
        else:
            level = "below min"
        if oper(setpoint, limit_value):
            msg = (
                f"{self.thermostat_type} zone {self.zone_name}: "
                f"scheduled {label.upper()} set point "
                f"({util.temp_value_with_units(setpoint)}) is {level} "
                f"limit ({util.temp_value_with_units(limit_value)})"
            )
            util.log_msg(f"WARNING: {msg}", mode=util.BOTH_LOG)
            eml.send_email_alert(subject=msg, body=f"{util.get_function_name()}: {msg}")
            return True
        else:
            return False

    def is_heat_mode(self) -> int:
        """Return 1 if in heat mode."""
        return int(
            (
                self.get_system_switch_position()
                == self.system_switch_position[self.HEAT_MODE]
            )
        )

    def is_cool_mode(self) -> int:
        """Return 1 if in cool mode."""
        return int(
            (
                self.get_system_switch_position()
                == self.system_switch_position[self.COOL_MODE]
            )
        )

    def is_dry_mode(self) -> int:
        """Return 1 if in dry mode."""
        return int(
            (
                self.get_system_switch_position()
                == self.system_switch_position[self.DRY_MODE]
            )
        )

    def is_auto_mode(self) -> int:
        """Return 1 if in auto mode."""
        return int(
            (
                self.get_system_switch_position()
                == self.system_switch_position[self.AUTO_MODE]
            )
        )

    def is_eco_mode(self) -> int:
        """Return 1 if in eco mode."""
        return int(
            (
                self.get_system_switch_position()
                == self.system_switch_position[self.ECO_MODE]
            )
        )

    def is_fan_mode(self) -> int:
        """Return 1 if fan mode enabled, else 0."""
        return (
            self.get_system_switch_position()
            == self.system_switch_position[self.FAN_MODE]
        )

    def is_off_mode(self) -> int:
        """Return 1 if fan mode enabled, else 0."""
        return (
            self.get_system_switch_position()
            == self.system_switch_position[self.OFF_MODE]
        )

    def is_controlled_mode(self) -> int:
        """Return True if mode is being controlled."""
        return self.current_mode in self.controlled_modes

    def is_heating(self) -> int:  # noqa R0201
        """Return 1 if heating relay is active, else 0."""
        return util.BOGUS_INT

    def is_cooling(self) -> int:  # noqa R0201
        """Return 1 if cooling relay is active, else 0."""
        return util.BOGUS_INT

    def is_drying(self) -> int:  # noqa R0201
        """Return 1 if drying relay is active, else 0."""
        return util.BOGUS_INT

    def is_auto(self) -> int:  # noqa R0201
        """Return 1 if auto relay is active, else 0."""
        return util.BOGUS_INT

    def is_eco(self) -> int:  # noqa R0201
        """Return 1 if eco relay is active, else 0."""
        return util.BOGUS_INT

    def is_fanning(self) -> int:  # noqa R0201
        """Return 1 if fan relay is active, else 0."""
        return util.BOGUS_INT

    def is_defrosting(self) -> int:  # noqa R0201
        """Return 1 if defrosting relay is active, else 0."""
        return util.BOGUS_INT

    def is_standby(self) -> int:  # noqa R0201
        """Return 1 if standby relay is active, else 0."""
        return util.BOGUS_INT

    def set_heat_setpoint(self, temp: int) -> None:  # noqa R0201
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in °F.
        returns:
            None
        """
        del temp
        util.log_msg(
            f"WARNING: {util.get_function_name()}: function is not yet "
            f"implemented on this thermostat, doing nothing",
            mode=util.BOTH_LOG,
            func_name=1,
        )

    def set_cool_setpoint(self, temp: int) -> None:  # noqa R0201
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in °F.
        returns:
            None
        """
        del temp
        util.log_msg(
            f"WARNING: {util.get_function_name()}: function is not yet "
            f"implemented on this thermostat, doing nothing",
            mode=util.BOTH_LOG,
            func_name=1,
        )

    def is_heat_deviation(self) -> bool:
        """
        Return True if heat is deviated.

        inputs:
            None
        returns:
            (bool): True if deviation exists.
        """
        return self.is_heat_mode() and self.is_temp_deviated_from_schedule()

    def is_cool_deviation(self) -> bool:
        """
        Return True if cool is deviated.

        inputs:
            None
        returns:
            (bool): True if deviation exists.
        """
        return self.is_cool_mode() and self.is_temp_deviated_from_schedule()

    # Thermostat-specific methods will be overloaded
    def get_display_temp(self) -> float:  # noqa R0201
        """Return the displayed temperature."""
        return float(util.BOGUS_INT)  # placeholder

    def get_display_humidity(self) -> Union[float, None]:  # noqa R0201
        """Return the displayed humidity."""
        return float(util.BOGUS_INT)  # placeholder

    def get_is_humidity_supported(self) -> bool:  # noqa R0201
        """Return humidity sensor status."""
        return util.BOGUS_BOOL  # placeholder

    def get_system_switch_position(self) -> int:  # noqa R0201
        """Return the 'SystemSwitchPosition'
        'SystemSwitchPosition' = 1 for heat, 2 for off
        """
        return util.BOGUS_INT  # placeholder

    def get_heat_setpoint_raw(self) -> float:  # noqa R0201
        """Return raw heat set point(number only, no units)."""
        return float(util.BOGUS_INT)  # placeholder

    def get_heat_setpoint(self) -> str:  # noqa R0201
        """Return raw heat set point(number and units)."""
        return util.BOGUS_STR  # placeholder

    def get_schedule_program_heat(self) -> dict:  # noqa R0201
        """
        Return the heat setpoint schedule.

        inputs:
            None
        returns:
            (dict): scheduled heat set points and times in °F.
        """
        return util.bogus_dict  # placeholder

    def get_schedule_heat_sp(self) -> float:
        """Return the heat setpoint."""
        return float(util.BOGUS_INT)  # placeholder

    def get_cool_setpoint_raw(self) -> float:  # noqa R0201
        """Return raw cool set point (number only, no units)."""
        return float(util.BOGUS_INT)  # placeholder

    def get_cool_setpoint(self) -> str:  # noqa R0201
        """Return raw cool set point (number and units)."""
        return util.BOGUS_STR  # placeholder

    def get_schedule_program_cool(self) -> dict:  # noqa R0201
        """
        Return the cool setpoint schedule.

        inputs:
            None
        returns:
            (dict): scheduled cool set points and times in °F.
        """
        return util.bogus_dict  # placeholder

    def get_schedule_cool_sp(self) -> float:
        """Return the cool setpoint."""
        return float(util.BOGUS_INT)  # placeholder

    def get_vacation_hold(self) -> bool:  # noqa R0201
        """Return True if thermostat is in vacation hold mode."""
        return util.BOGUS_BOOL  # placeholder

    def get_is_invacation_hold_mode(self) -> bool:  # noqa R0201
        """Return the 'IsInVacationHoldMode' setting."""
        return util.BOGUS_BOOL  # placeholder

    def get_temporary_hold_until_time(self) -> int:  # noqa R0201
        """Return the 'TemporaryHoldUntilTime'"""
        return util.BOGUS_INT  # placeholder

    def get_wifi_strength(self) -> float:  # noqa R0201
        """Return the wifi signal strength in dBm."""
        return float(util.BOGUS_INT)  # placeholder

    def get_wifi_status(self) -> bool:  # noqa R0201
        """Return the wifi connection status."""
        return util.BOGUS_BOOL  # placeholder

    def get_battery_voltage(self) -> float:  # noqa R0201
        """Return the battery voltage in volts."""
        return float(util.BOGUS_INT)  # placeholder

    def get_battery_status(self) -> bool:  # noqa R0201
        """Return the battery status."""
        return util.BOGUS_BOOL  # placeholder

    def refresh_zone_info(self, force_refresh=False) -> None:
        """
        Refresh zone info.

        inputs:
            force_refresh(bool): if True, ignore expiration timer.
        returns:
            None, cached data is refreshed.
        """
        now_time = time.time()
        # refresh if past expiration date or force_refresh option
        if force_refresh or (
            now_time >= (self.last_fetch_time + self.fetch_interval_sec)
        ):
            self.zone_info = {}
            self.last_fetch_time = now_time

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(
            f"display temp={util.temp_value_with_units(self.get_display_temp())}",
            mode=util.BOTH_LOG,
            func_name=1,
        )

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}", mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}", mode=util.BOTH_LOG
            )
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}", mode=util.BOTH_LOG
            )

        # cooling status
        if switch_position == self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}", mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}", mode=util.BOTH_LOG
            )
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}", mode=util.BOTH_LOG
            )

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG,
        )
        util.log_msg(f"vacation hold={self.get_vacation_hold()}", mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG,
        )
        util.log_msg(
            f"temporary hold until time={self.get_temporary_hold_until_time()}",
            mode=util.BOTH_LOG,
        )

    def get_vacation_hold_until_time(self) -> int:  # noqa R0201
        """
        Return the 'VacationHoldUntilTime'.

        inputs:
            None
        returns:
            (int): vacation hold time in minutes.
        """
        return util.BOGUS_INT  # not implemented

    def update_runtime_parameters(self):
        """use runtime parameter overrides.

        inputs:
            None
        returns:
            None, updates class variables.
        """
        # map user input keys to class members.
        # "thermostat_type is not overwritten
        user_input_to_class_mapping = {
            api.input_flds.thermostat_type: "thermostat_type",
            api.input_flds.zone: "zone_number",
            api.input_flds.poll_time: "poll_time_sec",
            api.input_flds.connection_time: "connection_time_sec",
            api.input_flds.tolerance: "tolerance_degrees",
            api.input_flds.target_mode: "target_mode",
            api.input_flds.measurements: "measurements",
        }

        if self.verbose:
            print("\n")
            util.log_msg(
                "supervisor runtime parameters:", mode=util.BOTH_LOG, func_name=1
            )
        for inp, cls_method in user_input_to_class_mapping.items():
            user_input = api.uip.get_user_inputs(api.uip.zone_name, inp)
            if user_input is not None:
                setattr(self, cls_method, user_input)
                if self.verbose:
                    # Add units to specific parameters for clarity
                    if inp == api.input_flds.poll_time:
                        display_value = f"{user_input} seconds"
                    elif inp == api.input_flds.connection_time:
                        display_value = f"{user_input} seconds"
                    elif inp == api.input_flds.tolerance:
                        display_value = util.temp_value_with_units(
                            user_input, precision=0
                        )
                    else:
                        display_value = str(user_input)
                    util.log_msg(
                        f"{inp}={display_value}", mode=util.BOTH_LOG, func_name=1
                    )

    def verify_current_mode(self, target_mode):
        """
        Verify current mode matches target mode.

        inputs:
            target_mode(str): target mode override
        returns:
            (bool): True if current mode matches target mode,
                    or target mode is not specified.
        """
        if target_mode is None:
            return True
        else:
            return bool(self.current_mode == target_mode)

    def revert_thermostat_mode(self, target_mode):
        """
        Revert thermostat mode to target mode.

        inputs:
            target_mode(str): target mode override
        returns:
            target_mode(str) target_mode, which may get updated by
            this function.
        """
        # do not switch directly from hot to cold
        if self.current_mode in self.heat_modes and target_mode in self.cool_modes:
            util.log_msg(
                f"WARNING: target mode={target_mode}, switching from "
                f"{self.current_mode} to OFF_MODE to prevent damage to HVAC",
                mode=util.BOTH_LOG,
                func_name=1,
            )
            target_mode = self.OFF_MODE

        # do not switch directly from cold to hot
        elif self.current_mode in self.cool_modes and target_mode in self.heat_modes:
            util.log_msg(
                f"WARNING: target mode={target_mode}, switching from "
                f"{self.current_mode} to OFF_MODE to prevent damage to HVAC",
                mode=util.BOTH_LOG,
                func_name=1,
            )
            target_mode = self.OFF_MODE

        # revert the mode to target
        # UNKNOWN mode is bypassed
        if target_mode == self.UNKNOWN_MODE and self.verbose:
            print(
                f"{util.get_function_name()}: target_mode='{target_mode}', "
                "doing nothing."
            )
        else:
            self.set_mode(target_mode)

        return target_mode

    def measure_thermostat_repeatability(
        self,
        measurements=30,
        poll_interval_sec=0,
        func=None,
        measure_response_time=False,
    ):
        """
        Measure Thermostat repeatability and report statistics.

        inputs:
            measurements(int): number of measurements
            poll_interval_sec(int): delay between measurements
            func(obj): target function to run during repeatabilty measurement.
            measure_response_time(bool): measure stats on response time instead
                                         of function.
        returns:
            (dict): measurement statistics.
        """
        data_lst = []
        stats = {}
        valid_datatypes = (int, float)
        # set default measurement method if not provided.
        if func is None:
            func = self.get_display_temp

        # title message
        if poll_interval_sec > 0.0:
            delay_msg = f" with {poll_interval_sec} sec. delay between measurements"
        else:
            delay_msg = ""
        print(
            f"\nThermostat response times for {measurements} measurements{delay_msg}..."
        )

        # measurement loop
        for measurement in range(measurements):
            t_start = time.time()
            data = func()  # target command
            t_end = time.time()

            # accumulate stats
            tdelta = t_end - t_start
            if measure_response_time:
                data_lst.append(tdelta)
                measurement_display = tdelta
            else:
                # check for valid return type
                if not isinstance(data, valid_datatypes):
                    raise TypeError(
                        f"metric value={data}, data type={type(data)}, the "
                        "metric repeatability assessment requires a "
                        "function that returns the metric of types "
                        f"{valid_datatypes}"
                    )
                data_lst.append(data)
                measurement_display = data
            util.log_msg(
                f"measurement {measurement}={measurement_display} "
                f"(deltaTime={tdelta:.3f} sec, delay={poll_interval_sec} sec)",
                mode=util.BOTH_LOG,
                func_name=1,
            )
            time.sleep(poll_interval_sec)

        # calc stats
        stats["measurements"] = measurements
        stats["mean"] = round(statistics.mean(data_lst), 2)
        stats["stdev"] = round(statistics.stdev(data_lst), 2)
        stats["min"] = round(min(data_lst), 2)
        stats["max"] = round(max(data_lst), 2)
        stats["3sigma_upper"] = round((3.0 * stats["stdev"] + stats["mean"]), 2)
        stats["6sigma_upper"] = round((6.0 * stats["stdev"] + stats["mean"]), 2)
        return stats

    def display_basic_thermostat_summary(self, mode=util.STDOUT_LOG):
        """
        Display basic thermostat summary.

        inputs:
            mode(int): target log for data.
        returns:
            None, prints data to log and/or console.
        """
        print("\n")
        util.log_msg("current thermostat settings...", mode=mode, func_name=1)
        util.log_msg(f"zone name='{self.zone_name}'", mode=mode, func_name=1)
        sw_pos = util.get_key_from_value(
            self.system_switch_position, self.get_system_switch_position()
        )
        util.log_msg(
            f"system switch position: {self.get_system_switch_position()} "
            f"({sw_pos})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"thermostat display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"thermostat display humidity="
            f"{util.humidity_value_with_units(self.get_display_humidity())}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"heat set point={self.get_heat_setpoint()}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"cool set point={self.get_cool_setpoint()}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"heat schedule set point="
            f"{util.temp_value_with_units(self.get_schedule_heat_sp())}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"cool schedule set point="
            f"{util.temp_value_with_units(self.get_schedule_cool_sp())}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"(schedule) heat program={self.get_schedule_program_heat()}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"(schedule) cool program={self.get_schedule_program_cool()}",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"heat mode={self.is_heat_mode()} "
            f"(actively heating={self.is_heating()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"cool mode={self.is_cool_mode()} "
            f"(actively cooling={self.is_cooling()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"dry mode={self.is_dry_mode()} (actively drying={self.is_drying()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"auto mode={self.is_auto_mode()} (actively auto={self.is_auto()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"eco mode={self.is_eco_mode()} (actively eco={self.is_eco()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(
            f"fan mode={self.is_fan_mode()} (actively fanning={self.is_fanning()})",
            mode=mode,
            func_name=1,
        )
        util.log_msg(f"off mode={self.is_off_mode()}", mode=mode, func_name=1)
        util.log_msg(f"hold={self.get_vacation_hold()}", mode=mode, func_name=1)
        util.log_msg(
            f"temporary hold minutes={self.get_temporary_hold_until_time()}",
            mode=mode,
            func_name=1,
        )

    def revert_temperature_deviation(self, setpoint=None, msg=""):
        """
        Revert the temperature deviation.

        inputs:
            setpoint(int or float): setpoint in °F
            msg(str): status message to display.
        returns:
            None
        """
        if setpoint is None:
            setpoint = self.current_setpoint

        eml.send_email_alert(
            subject=f"{self.thermostat_type} {self.current_mode.upper()} "
            f"deviation alert on zone {self.zone_name}",
            body=msg,
        )
        util.log_msg(
            f"\n*** {self.thermostat_type} {self.current_mode.upper()} "
            f"deviation detected on zone {self.zone_name}, "
            f"reverting thermostat to heat schedule ***\n",
            mode=util.BOTH_LOG,
        )
        self.revert_setpoint_func(setpoint)

    def function_not_supported(self, *_, **__):
        """Function for unsupported activities."""
        util.log_msg(
            f"WARNING (in {util.get_function_name(2)}): function call is "
            f"not supported on this thermostat type",
            mode=util.BOTH_LOG,
        )

    def display_runtime_settings(self):
        """
        Display runtime settings to console.

        inputs:
            None
        returns:
            None
        """
        # poll time setting:
        util.log_msg(
            f"polling time set to {self.poll_time_sec / 60.0:.1f} minutes "
            f"({self.poll_time_sec} seconds)",
            mode=util.BOTH_LOG,
        )

        # reconnection time to thermostat server:
        util.log_msg(
            f"server re-connect time set to "
            f"{int(self.connection_time_sec / 60.0)} minutes "
            f"({self.connection_time_sec} seconds)",
            mode=util.BOTH_LOG,
        )

        # tolerance to set point:
        util.log_msg(
            f"tolerance to set point is set to "
            f"{util.temp_value_with_units(self.tolerance_degrees, precision=0)}",
            mode=util.BOTH_LOG,
        )

    def display_session_settings(self):
        """
        Display session settings to console.

        inputs:
            None
        returns:
            None
        """
        # set log file name
        util.log_msg.file_name = (
            self.thermostat_type + "_" + str(self.zone_name) + ".txt"
        )

        util.log_msg(
            f"{self.thermostat_type} thermostat zone {self.zone_name} "
            f"monitoring service\n",
            mode=util.BOTH_LOG,
        )

        util.log_msg("session settings:", mode=util.BOTH_LOG)

        util.log_msg(
            "thermostat %s for %s\n"
            % (
                ["is being monitored", "will be reverted"][self.revert_deviations],
                [
                    "energy consuming deviations\n("
                    "e.g. heat setpoint above schedule "
                    "setpoint, cool setpoint below schedule"
                    " setpoint)",
                    "all schedule deviations",
                ][self.revert_all_deviations],
            ),
            mode=util.BOTH_LOG,
        )

    def supervisor_loop(self, Thermostat, session_count, measurement, debug):
        """
        Loop through supervisor algorithm.

        inputs:
            Thermostat(obj):  Thermostat instance object
            session_count(int):  current session
            measurement(int):  current measurement index
            debug(bool): debug flag
        returns:
            measurement(int): current measurement count
        """
        global connection_ok  # noqa W603

        # initialize poll counter
        poll_count = 1
        previous_mode_dict = {}

        # Calculate maximum loop time based on expected measurements
        # Allow enough time for all measurements plus network operations
        max_measurements = api.uip.get_user_inputs(
            api.uip.zone_name, api.input_flds.measurements
        )
        if max_measurements:
            # Calculate max time: (measurements * poll_time) + generous buffer
            # for network operations and retries
            max_loop_time_sec = (max_measurements * self.poll_time_sec) + (
                max_measurements * 300
            )  # 5 min buffer per measurement
            loop_start_time = time.time()
            util.log_msg(
                f"supervisor_loop: max_loop_time="
                f"{max_loop_time_sec / 86400.0:.1f} days "
                f"for {max_measurements} measurements",
                mode=util.DUAL_STREAM_LOG,
                func_name=1,
            )
        else:
            max_loop_time_sec = None
            loop_start_time = None

        # poll thermostat settings
        while not api.uip.max_measurement_count_exceeded(measurement):
            # Check for overall loop timeout to prevent indefinite hanging
            # This check must happen BEFORE potentially blocking operations
            if max_loop_time_sec and loop_start_time:
                elapsed_time = time.time() - loop_start_time
                if elapsed_time > max_loop_time_sec:
                    util.log_msg(
                        f"supervisor_loop: exceeded max loop time "
                        f"({elapsed_time:.1f}s > {max_loop_time_sec}s), "
                        f"exiting loop at measurement {measurement}",
                        mode=util.BOTH_LOG,
                        func_name=1,
                    )
                    # Set measurement to exceed max to exit outer loop in
                    # supervise.py The outer loop will terminate when
                    # measurement > max_measurements
                    measurement = max_measurements + 1
                    break

            # query thermostat for current settings and set points
            # Record start time for this iteration
            iteration_start_time = time.time()
            # Wrap in try/except to catch any unexpected hangs/exceptions
            try:
                current_mode_dict = self.get_current_mode(
                    session_count,
                    poll_count,
                    flag_all_deviations=self.revert_all_deviations,
                )
            except Exception as e:
                util.log_msg(
                    f"supervisor_loop: exception in get_current_mode: {e}",
                    mode=util.BOTH_LOG,
                    func_name=1,
                )
                # Check if we've exceeded max time due to retries
                if max_loop_time_sec and loop_start_time:
                    elapsed_time = time.time() - loop_start_time
                    if elapsed_time > max_loop_time_sec:
                        util.log_msg(
                            f"supervisor_loop: exceeded max loop time after "
                            f"exception ({elapsed_time:.1f}s > "
                            f"{max_loop_time_sec}s), exiting loop",
                            mode=util.BOTH_LOG,
                            func_name=1,
                        )
                        # Set measurement to exceed max to exit outer loop
                        measurement = max_measurements + 1
                        break
                # Re-raise to maintain existing error handling behavior
                raise

            # Check if get_current_mode took too long
            iteration_elapsed = time.time() - iteration_start_time
            if max_loop_time_sec and iteration_elapsed > (
                max_loop_time_sec / max_measurements
            ):
                util.log_msg(
                    f"supervisor_loop: single iteration took {iteration_elapsed:.1f}s "
                    f"(exceeds {max_loop_time_sec / max_measurements:.1f}s threshold), "
                    "may indicate network issues",
                    mode=util.BOTH_LOG,
                    func_name=1,
                )

            # debug data on change from previous poll
            # note this check is probably hyper-sensitive, since status msg
            # change could trigger this extra report.
            if current_mode_dict != previous_mode_dict:
                if debug:
                    self.report_heating_parameters()
                previous_mode_dict = current_mode_dict  # latch

            # revert thermostat mode if not matching target
            if not self.verify_current_mode(
                api.uip.get_user_inputs(api.uip.zone_name, api.input_flds.target_mode)
            ):
                api.uip.set_user_inputs(
                    api.uip.zone_name,
                    api.input_flds.target_mode,
                    self.revert_thermostat_mode(
                        api.uip.get_user_inputs(
                            api.uip.zone_name, api.input_flds.target_mode
                        )
                    ),
                )

            # revert thermostat to schedule if heat override is detected
            if (
                self.revert_deviations
                and self.is_controlled_mode()
                and self.is_temp_deviated_from_schedule()
            ):
                self.revert_temperature_deviation(
                    self.schedule_setpoint, current_mode_dict["status_msg"]
                )

            # increment poll count
            poll_count += 1
            measurement += 1

            # polling delay
            time.sleep(self.poll_time_sec)

            # refresh zone info
            connection_ok = True
            self.refresh_zone_info()

            # reconnect
            if (
                (time.time() - self.session_start_time_sec) > self.connection_time_sec
            ) or not connection_ok:
                util.log_msg(
                    "forcing re-connection to thermostat...", mode=util.BOTH_LOG
                )
                del Thermostat
                break  # force reconnection

        # return measurement count
        return measurement


def create_thermostat_instance(
    thermostat_type, zone, ThermostatClass, ThermostatZone, verbose=True
):
    """
    Create Thermostat and Zone instances.

    inputs:
        tstat(int):  thermostat_type
        zone(int): zone number
        zone_name(str): name of zone
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
        verbose(bool): debug flag
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = verbose  # debug mode set

    # verify required env vars
    api.verify_required_env_variables(thermostat_type, str(zone), verbose=verbose)

    # import hardware module
    api.load_hardware_library(thermostat_type)

    # create Thermostat object
    Thermostat = ThermostatClass(zone, verbose=verbose)
    if verbose:
        Thermostat.print_all_thermostat_metadata(zone)

    # create Zone object
    Zone = ThermostatZone(Thermostat, verbose=verbose)

    # update runtime overrides
    # thermostat_type
    api.uip.set_user_inputs(
        api.uip.zone_name, api.input_flds.thermostat_type, thermostat_type
    )
    # zone
    api.uip.set_user_inputs(api.uip.zone_name, api.input_flds.zone, zone)
    Zone.update_runtime_parameters()

    return Thermostat, Zone


def thermostat_basic_checkout(thermostat_type, zone, ThermostatClass, ThermostatZone):
    """
    Perform basic Thermostat checkout.

    inputs:
        tstat(int):  thermostat_type
        zone(int): zone number
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = True  # debug mode set

    # create class instances
    Thermostat, Zone = create_thermostat_instance(
        thermostat_type, zone, ThermostatClass, ThermostatZone
    )

    Zone.display_basic_thermostat_summary()

    return Thermostat, Zone


def get_wifi_status_display(wifi_status):
    """
    Return string to display based on wifi status.

    inputs:
        wifi_status(bool, float, None): wifi status.
    returns:
        (str): wifi status
    """
    wifi_status_display_dict = {
        False: "weak",
        True: "ok",
    }
    try:
        return wifi_status_display_dict[wifi_status]
    except KeyError:
        return "unknown"


def get_battery_status_display(battery_status):
    """
    Return string to display based on battery status.

    inputs:
        wifi_status(bool, float, None): wifi status.
    returns:
        (str): wifi status
    """
    battery_status_display_dict = {
        False: "bad",
        True: "ok",
    }
    try:
        return battery_status_display_dict[battery_status]
    except KeyError:
        return "unknown"


def print_select_data_from_all_zones(
    thermostat_type,
    zone_lst,
    ThermostatClass,
    ThermostatZone,
    display_wifi=True,
    display_battery=True,
    display_outdoor_weather=True,
):
    """
    Cycle through all zones and print out select data.

    inputs:
        tstat(int):  thermostat_type
        zone_lst(list): list of zones
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
        display_wifi(bool): display wifi status
        display_battery(bool): display battery status
        display_outdoor_weather(bool): display outdoor weather data
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = False  # debug mode unset
    print("\nquerying select data for all zones:")

    # Get outdoor weather data once if enabled (same for all zones)
    outdoor_weather_data = None
    if display_outdoor_weather:
        try:
            # Get zip code from thermostat configuration
            zip_code = api.SUPPORTED_THERMOSTATS.get(thermostat_type, {}).get(
                "zip_code"
            )
            if zip_code:
                api_key = weather.get_weather_api_key()
                outdoor_weather_data = weather.get_outdoor_weather(zip_code, api_key)
        except Exception as e:
            util.log_msg(
                f"Failed to get outdoor weather data: {e}",
                mode=util.BOTH_LOG,
                func_name=1,
            )

    for zone in zone_lst:
        # create class instances
        Thermostat, Zone = create_thermostat_instance(
            thermostat_type, zone, ThermostatClass, ThermostatZone, verbose=False
        )
        # zone temperature
        display_temp = Zone.get_display_temp()
        msg = f"zone: {zone}, name: {Zone.zone_name}, temp: {display_temp:.1f} °F"

        # zone wifi strength
        if display_wifi:
            wifi_strength = Zone.get_wifi_strength()
            wifi_status = Zone.get_wifi_status()
            wifi_status_display = get_wifi_status_display(wifi_status)
            msg += f", wifi strength: {wifi_strength} dBm ({wifi_status_display})"

        # zone battery stats
        if display_battery:
            battery_voltage = Zone.get_battery_voltage()
            battery_status = Zone.get_battery_status()
            battery_status_display = get_battery_status_display(battery_status)
            msg += (
                f", battery voltage: {battery_voltage:.2f} volts "
                f"({battery_status_display})"
            )

        # outdoor weather data
        if display_outdoor_weather and outdoor_weather_data:
            weather_display = weather.format_weather_display(outdoor_weather_data)
            msg += f", {weather_display}"

        print(msg)

    return Thermostat, Zone


class AuthenticationError(ValueError):
    """denoted if we are completely unable to authenticate."""

    pass



"""
Integration test module for emulator.py.

This test requires connection to emulator thermostat.
"""

# built-in imports
import math
import unittest

# local imports
from thermostatsupervisor import emulator
from thermostatsupervisor import emulator_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in emulator.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # emulator argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "emulator",  # thermostat type
            "0",  # zone
            "5",  # poll time in sec, this value violates min
            # cycle time for TCC if reverting temperature deviation
            "11",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "6",  # number of measurements
        ]
        self.mod = emulator
        self.mod_config = emulator_config


class FunctionalIntegrationTest(IntegrationTest, utc.FunctionalIntegrationTest):
    """
    Test functional performance of emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.trait_field = None
        self.metadata_field = "display_temp"
        self.metadata_type = float

    def test_deviation_functionality(self):
        """Test emulator deviation functionality for improved testability."""
        self.print_test_name()

        # Create thermostat zone instance
        Thermostat = self.mod.ThermostatClass(zone="0", verbose=False)
        Zone = self.mod.ThermostatZone(Thermostat, verbose=False)

        try:
            # Test 1: Initial state - no deviation data
            self.assertFalse(
                Zone.has_deviation_data(), "Should have no deviation data initially"
            )

            # Get baseline values (stored for potential future use)
            # Note: These values are not currently used in assertions
            # due to random noise

            # Test 2: Create deviation file
            Zone.create_deviation_file()
            self.assertTrue(
                Zone.has_deviation_data(), "Should have deviation file after creation"
            )

            # Test 3: Set specific deviation values
            test_temp = 85.5
            test_humidity = 60.0
            test_heat_sp = 68.0
            test_cool_sp = 76.0

            Zone.set_deviation_value("display_temp", test_temp)
            Zone.set_deviation_value("display_humidity", test_humidity)
            Zone.set_deviation_value("heat_setpoint", test_heat_sp)
            Zone.set_deviation_value("cool_setpoint", test_cool_sp)

            # Test 4: Verify deviation values are returned
            self.assertTrue(
                Zone.has_deviation_data("display_temp"), "Should have temp deviation"
            )
            self.assertTrue(
                Zone.has_deviation_data("display_humidity"),
                "Should have humidity deviation",
            )
            self.assertTrue(
                Zone.has_deviation_data("heat_setpoint"),
                "Should have heat setpoint deviation",
            )
            self.assertTrue(
                Zone.has_deviation_data("cool_setpoint"),
                "Should have cool setpoint deviation",
            )

            self.assertEqual(
                Zone.get_display_temp(), test_temp, "Should return deviated temperature"
            )
            self.assertEqual(
                Zone.get_display_humidity(),
                test_humidity,
                "Should return deviated humidity",
            )
            self.assertEqual(
                Zone.get_heat_setpoint_raw(),
                test_heat_sp,
                "Should return deviated heat setpoint",
            )
            self.assertEqual(
                Zone.get_cool_setpoint_raw(),
                test_cool_sp,
                "Should return deviated cool setpoint",
            )

            # Test 5: Test individual value retrieval
            self.assertEqual(
                Zone.get_deviation_value("display_temp"),
                test_temp,
                "Should retrieve specific deviation value",
            )
            self.assertIsNone(
                Zone.get_deviation_value("nonexistent_key"),
                "Should return None for missing key",
            )
            self.assertEqual(
                Zone.get_deviation_value("nonexistent_key", "default"),
                "default",
                "Should return default for missing key",
            )

            # Test 6: Clear deviation data
            Zone.clear_deviation_data()
            self.assertFalse(
                Zone.has_deviation_data(),
                "Should have no deviation data after clearing",
            )

            # Test 7: Verify normal behavior is restored
            # Note: We can't compare exact values due to random noise,
            # but values should be in normal range
            current_temp = Zone.get_display_temp()
            current_humidity = Zone.get_display_humidity()

            # Values should be within normal variation range from base values
            temp_diff = abs(current_temp - Zone.get_parameter("display_temp"))
            humidity_diff = abs(
                current_humidity - Zone.get_parameter("display_humidity")
            )

            self.assertLessEqual(
                temp_diff,
                emulator_config.NORMAL_TEMP_VARIATION,
                "Temperature should be within normal variation after clearing "
                "deviation",
            )
            self.assertLessEqual(
                humidity_diff,
                emulator_config.NORMAL_HUMIDITY_VARIATION,
                "Humidity should be within normal variation after clearing deviation",
            )

        finally:
            # Cleanup: ensure deviation file is removed
            Zone.clear_deviation_data()


class SuperviseIntegrationTest(IntegrationTest, utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


class PerformanceIntegrationTest(IntegrationTest, utc.PerformanceIntegrationTest):
    """
    Test performance of emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = 1
        self.timing_measurements = 30  # fast measurement

        # temperature and humidity repeatability measurements
        # data is uniform distribution, 1 std= (range)/sqrt(12)
        # adding 0.1 buffer to prevent false failures due to rounding, etc.
        self.temp_stdev_limit = (
            emulator_config.NORMAL_TEMP_VARIATION * 2 / math.sqrt(12) + 0.1
        )
        self.temp_repeatability_measurements = 100  # number of temp msmts.
        self.humidity_stdev_limit = (
            emulator_config.NORMAL_HUMIDITY_VARIATION * 2 / math.sqrt(12) + 0.1
        )
        self.humidity_repeatability_measurements = 100  # number of temp msmts.
        self.poll_interval_sec = 0.5  # delay between repeatability msmts


class EmulatorUnitTest(utc.UnitTest):
    """Unit tests for specific emulator methods."""

    def setUp(self):
        super().setUp()
        self.setup_mock_thermostat_zone()

        # Create emulator instance
        self.thermostat = emulator.ThermostatClass(zone="0", verbose=False)
        self.zone = emulator.ThermostatZone(self.thermostat, verbose=False)
        self.zone.update_runtime_parameters()

    def tearDown(self):
        self.teardown_mock_thermostat_zone()
        super().tearDown()

    def test_is_dry_mode(self):
        """Test is_dry_mode() method."""
        from unittest.mock import patch

        # Test when dry mode is enabled
        with patch.object(self.zone, "get_system_switch_position") as mock_switch:
            mock_switch.return_value = self.zone.system_switch_position[
                self.zone.DRY_MODE
            ]
            result = self.zone.is_dry_mode()
            self.assertEqual(result, 1)

        # Test when dry mode is disabled
        with patch.object(self.zone, "get_system_switch_position") as mock_switch:
            mock_switch.return_value = self.zone.system_switch_position[
                self.zone.HEAT_MODE
            ]
            result = self.zone.is_dry_mode()
            self.assertEqual(result, 0)

    def test_is_defrosting(self):
        """Test is_defrosting() method."""
        from unittest.mock import patch

        # Test when defrosting is active
        with patch.object(self.zone, "refresh_zone_info"), patch.object(
            self.zone, "get_parameter"
        ) as mock_param:
            mock_param.return_value = True
            result = self.zone.is_defrosting()
            self.assertEqual(result, 1)

        # Test when defrosting is not active
        with patch.object(self.zone, "refresh_zone_info"), patch.object(
            self.zone, "get_parameter"
        ) as mock_param:
            mock_param.return_value = False
            result = self.zone.is_defrosting()
            self.assertEqual(result, 0)


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Unit test for SHT31 environment variable fallback logic.

This test validates the fix for issue where SHT31 ThermostatClass
was failing when environment variables are missing in unit test mode.
"""

import os
import unittest
from unittest.mock import patch

from thermostatsupervisor import sht31
from thermostatsupervisor import sht31_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class TestSHT31MissingEnvVar(utc.UnitTest):
    """Test SHT31 environment variable fallback logic."""

    def setUp(self):
        """Set up test fixtures."""
        super().setUp()
        self.original_unit_test_mode = util.unit_test_mode

    def tearDown(self):
        """Clean up after tests."""
        util.unit_test_mode = self.original_unit_test_mode
        super().tearDown()

    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    @patch.dict(os.environ, {}, clear=False)
    def test_missing_env_var_fallback_in_unit_test_mode(self, mock_spawn):
        """
        Test that missing environment variables fall back to localhost in unit
        test mode.

        This reproduces and fixes the exact error scenario:
        - Zone 1 (regular zone, not unit test zone 99)
        - Unit test mode enabled
        - Environment variable SHT31_REMOTE_IP_ADDRESS_1 is missing or placeholder
        - Should fall back to 127.0.0.1 instead of None or placeholder
        """
        mock_spawn.return_value = None

        # Save original state
        original_mode = util.unit_test_mode

        try:
            # Explicitly set unit test mode
            util.unit_test_mode = True

            # Verify unit test mode is actually set
            self.assertTrue(util.unit_test_mode,
                            "unit_test_mode should be True for this test")

            # Clear the env var from OS environment if it exists
            # to simulate missing environment variable scenario
            if 'SHT31_REMOTE_IP_ADDRESS_1' in os.environ:
                del os.environ['SHT31_REMOTE_IP_ADDRESS_1']

            # This should NOT fail even if env var is missing or has placeholder
            tstat = sht31.ThermostatClass(1, verbose=False)

            # Debug information to help diagnose failures
            if tstat.ip_address != "127.0.0.1":
                from thermostatsupervisor import environment as env
                env_result = env.get_env_variable('SHT31_REMOTE_IP_ADDRESS_1')
                self.fail(
                    f"Expected IP address '127.0.0.1' but got "
                    f"'{tstat.ip_address}'. "
                    f"util.unit_test_mode={util.unit_test_mode}, "
                    f"env_result={env_result}, "
                    f"util module id={id(util)}"
                )

            # Verify the IP address fallback works for both missing and
            # placeholder values
            self.assertEqual(tstat.ip_address, "127.0.0.1")

            # Verify the URL is properly formed
            expected_components = [
                "http://127.0.0.1:5000",
                "/unit",
                "measurements=10",
                "seed=127"
            ]
            for component in expected_components:
                self.assertIn(component, tstat.url)

            # Verify URL does NOT contain None or empty host
            self.assertNotIn("http://:5000", tstat.url)
            self.assertNotIn("None", tstat.url)
        finally:
            # Always restore original mode
            util.unit_test_mode = original_mode

    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    @patch.dict(os.environ, {}, clear=False)
    def test_missing_env_var_fails_in_non_unit_test_mode(self, mock_spawn):
        """
        Test that missing environment variables still cause ValueError in
        non-unit test mode.

        This ensures the fallback behavior is only active in unit test mode.
        """
        mock_spawn.return_value = None

        # Save original state
        original_mode = util.unit_test_mode

        try:
            # Explicitly disable unit test mode
            util.unit_test_mode = False

            # Verify unit test mode is actually disabled
            self.assertFalse(util.unit_test_mode,
                             "unit_test_mode should be False for this test")

            # Clear the env var from OS environment if it exists
            if 'SHT31_REMOTE_IP_ADDRESS_1' in os.environ:
                del os.environ['SHT31_REMOTE_IP_ADDRESS_1']

            # Create temporary environment where the env var is truly missing
            # by temporarily moving any supervisor-env.txt file
            import tempfile
            import shutil

            supervisor_env_path = "supervisor-env.txt"
            backup_path = None

            if os.path.exists(supervisor_env_path):
                # Use secure temporary file creation
                with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                    backup_path = temp_file.name
                shutil.move(supervisor_env_path, backup_path)

            try:
                # This should fail when env var is missing and not in unit
                # test mode
                with self.assertRaises(ValueError) as context:
                    sht31.ThermostatClass(1, verbose=False)

                # Should get a clear error message about the missing IP address
                error_msg = str(context.exception)
                self.assertIn("SHT31_REMOTE_IP_ADDRESS_1", error_msg)
                self.assertIn("empty or missing", error_msg)

            finally:
                # Restore supervisor-env.txt if it existed
                if backup_path and os.path.exists(backup_path):
                    shutil.move(backup_path, supervisor_env_path)

        finally:
            # Always restore original mode
            util.unit_test_mode = original_mode

    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    def test_placeholder_value_preserved_in_non_unit_test_mode(self, mock_spawn):
        """
        Test that placeholder values are preserved in non-unit-test mode.

        This ensures that in production, placeholder values don't get replaced.
        """
        mock_spawn.return_value = None
        util.unit_test_mode = False

        # Create a temporary supervisor-env.txt with placeholder value
        import tempfile

        original_cwd = os.getcwd()
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                os.chdir(temp_dir)

                # Create supervisor-env.txt with placeholder value
                with open("supervisor-env.txt", "w") as f:
                    f.write("SHT31_REMOTE_IP_ADDRESS_1=***\n")

                # Test that placeholder value is NOT replaced in non-unit-test
                # mode
                tstat = sht31.ThermostatClass(1, verbose=False)
                self.assertEqual(tstat.ip_address, "***")

        finally:
            os.chdir(original_cwd)

    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    @patch.dict(os.environ, {}, clear=True)  # Clear environment variables
    def test_placeholder_value_fallback_in_unit_test_mode(self, mock_spawn):
        """
        Test that placeholder values (like '***') fall back to localhost in
        unit test mode.

        This handles the specific CI scenario where supervisor-env.txt has
        placeholder values.
        """
        mock_spawn.return_value = None
        util.unit_test_mode = True

        # Create a temporary supervisor-env.txt with placeholder value
        import tempfile

        original_cwd = os.getcwd()
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                os.chdir(temp_dir)

                # Create supervisor-env.txt with placeholder value
                with open("supervisor-env.txt", "w") as f:
                    f.write("SHT31_REMOTE_IP_ADDRESS_1=***\n")

                # Test that placeholder value gets replaced with localhost
                tstat = sht31.ThermostatClass(1, verbose=False)
                self.assertEqual(tstat.ip_address, "127.0.0.1")

        finally:
            os.chdir(original_cwd)

    @patch.object(sht31.ThermostatClass, "spawn_flask_server")
    def test_unit_test_zone_still_works(self, mock_spawn):
        """Test that the unit test zone (99) still works as expected."""
        mock_spawn.return_value = None
        util.unit_test_mode = True

        # Zone 99 should work (this was already working before the fix)
        tstat = sht31.ThermostatClass(
            sht31_config.UNIT_TEST_ZONE,
            verbose=False
        )

        # Should get local IP via existing logic in environment.py
        self.assertIsNotNone(tstat.ip_address)
        self.assertIn("http://", tstat.url)
        self.assertNotIn("http://:5000", tstat.url)


if __name__ == "__main__":
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""
Unit test for nest OAuth token refresh functionality.

This test validates the OAuth token refresh fix without requiring
actual nest hardware connection or valid OAuth credentials.
"""
import json
import os
import tempfile
import unittest
from unittest.mock import Mock, patch

# third party imports
import oauthlib.oauth2.rfc6749.errors

# local imports
from thermostatsupervisor import nest
from tests import unit_test_common as utc


class TestNestOAuthRefresh(utc.UnitTest):
    """Test nest OAuth token refresh functionality."""

    def setUp(self):
        """Set up test fixtures."""
        super().setUp()
        self.print_test_name()

        # Create a temporary cache file
        self.test_cache_file = tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".json"
        )
        self.cache_file_path = self.test_cache_file.name

        # Sample initial token cache with valid tokens
        self.initial_token_data = {
            "access_token": "ya29.a0AfB_byOLD_ACCESS_TOKEN_EXPIRED",
            "refresh_token": (
                "1//040jmes6vvJ7gCgYIARAAGAQSNwF-L9IrdajcBypXM_Rbh0K"
                "clSy9j-oY3XYl8tb6Zhu1_0BgAPseN6pkAneuDmo0MY8qGGz8I"
                "ck"
            ),
            "expires_in": 3600,
        }

        # Write initial token data
        json.dump(self.initial_token_data, self.test_cache_file, indent=4)
        self.test_cache_file.close()

    def tearDown(self):
        """Clean up test fixtures."""
        if os.path.exists(self.cache_file_path):
            os.unlink(self.cache_file_path)

    @patch("thermostatsupervisor.nest.requests.post")
    def test_refresh_oauth_token_with_new_refresh_token(self, mock_post):
        """Test OAuth refresh when server returns a new refresh token."""
        # Mock successful refresh response with new refresh token
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "access_token": "ya29.a0AfB_byNEW_ACCESS_TOKEN",
            "refresh_token": "1//040NEW_REFRESH_TOKEN_DIFFERENT",
            "expires_in": 3600,
        }
        mock_post.return_value = mock_response

        # Create a minimal thermostat instance
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.client_id = "test_client_id"
        thermostat.client_secret = "test_client_secret"

        # Call the refresh method
        thermostat.refresh_oauth_token()

        # Verify the cache file was updated correctly
        with open(self.cache_file_path, "r", encoding="utf-8") as f:
            updated_data = json.load(f)

        # Check that both tokens were updated
        self.assertEqual(updated_data["access_token"], "ya29.a0AfB_byNEW_ACCESS_TOKEN")
        self.assertEqual(
            updated_data["refresh_token"], "1//040NEW_REFRESH_TOKEN_DIFFERENT"
        )
        self.assertEqual(updated_data["expires_in"], 3600)

        # Verify the refresh token was not overwritten with access token
        self.assertNotEqual(updated_data["refresh_token"], updated_data["access_token"])

    @patch("thermostatsupervisor.nest.requests.post")
    def test_refresh_oauth_token_without_new_refresh_token(self, mock_post):
        """Test OAuth refresh when server does NOT return a new refresh token."""
        # Mock successful refresh response without new refresh token
        # (common for Google OAuth)
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "access_token": "ya29.a0AfB_byNEW_ACCESS_TOKEN",
            "expires_in": 3600,
            # No refresh_token in response
        }
        mock_post.return_value = mock_response

        # Save original refresh token for comparison
        original_refresh_token = self.initial_token_data["refresh_token"]

        # Create a minimal thermostat instance
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.client_id = "test_client_id"
        thermostat.client_secret = "test_client_secret"

        # Call the refresh method
        thermostat.refresh_oauth_token()

        # Verify the cache file was updated correctly
        with open(self.cache_file_path, "r", encoding="utf-8") as f:
            updated_data = json.load(f)

        # Check that access token was updated but refresh token preserved
        self.assertEqual(updated_data["access_token"], "ya29.a0AfB_byNEW_ACCESS_TOKEN")
        self.assertEqual(updated_data["refresh_token"], original_refresh_token)
        self.assertEqual(updated_data["expires_in"], 3600)

        # Most importantly: verify refresh token was NOT overwritten with
        # access token
        self.assertNotEqual(updated_data["refresh_token"], updated_data["access_token"])

    @patch("thermostatsupervisor.nest.requests.post")
    def test_refresh_oauth_token_failure(self, mock_post):
        """Test OAuth refresh failure handling."""
        # Mock failed refresh response
        mock_response = Mock()
        mock_response.ok = False
        mock_response.status_code = 400
        mock_response.text = (
            '{"error": "invalid_grant", '
            '"error_description": "Token has been expired or '
            'revoked."}'
        )
        mock_post.return_value = mock_response

        # Create a minimal thermostat instance
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.client_id = "test_client_id"
        thermostat.client_secret = "test_client_secret"

        # Call the refresh method and expect exception
        with self.assertRaises(Exception) as context:
            thermostat.refresh_oauth_token()

        # Check that the exception contains the status code
        self.assertIn("Failed to refresh token: 400", str(context.exception))

        # Verify the cache file was NOT modified on failure
        with open(self.cache_file_path, "r", encoding="utf-8") as f:
            unchanged_data = json.load(f)

        # Data should be unchanged
        self.assertEqual(unchanged_data, self.initial_token_data)

    @patch("thermostatsupervisor.nest.requests.post")
    def test_get_device_data_with_invalid_grant_retry(self, mock_post):
        """
        Test get_device_data retries after token refresh on InvalidGrantError.
        """
        # Mock successful refresh response
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "access_token": "ya29.a0AfB_byNEW_REFRESHED_ACCESS_TOKEN",
            "expires_in": 3600,
        }
        mock_post.return_value = mock_response

        # Create a minimal thermostat instance
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.client_id = "test_client_id"
        thermostat.client_secret = "test_client_secret"
        thermostat.devices = []

        # Mock thermostat_obj with _client that needs token reload
        mock_tstat_obj = Mock()
        mock_client = Mock()
        mock_client.token = self.initial_token_data
        mock_tstat_obj._client = mock_client

        # First call raises InvalidGrantError, second succeeds
        mock_devices = [{"name": "test_device"}]
        mock_tstat_obj.get_devices.side_effect = [
            oauthlib.oauth2.rfc6749.errors.InvalidGrantError("invalid_grant"),
            mock_devices,
        ]
        thermostat.thermostat_obj = mock_tstat_obj

        # Call get_device_data - should catch error, refresh, and retry
        result = thermostat.get_device_data()

        # Verify refresh was called
        mock_post.assert_called_once()

        # Verify get_devices was called twice (initial + retry)
        self.assertEqual(mock_tstat_obj.get_devices.call_count, 2)

        # Verify token was reloaded into client
        self.assertIsNotNone(mock_client.token)
        self.assertEqual(
            mock_client.token["access_token"],
            "ya29.a0AfB_byNEW_REFRESHED_ACCESS_TOKEN",
        )

        # Verify devices were returned from retry
        self.assertEqual(result, mock_devices)


if __name__ == "__main__":
    unittest.main()



"""Test module to verify duplicate STDOUT message fixes."""

import os
import sys
import shutil
import unittest
from io import StringIO

# Add the project to path if needed  # noqa: E402
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from thermostatsupervisor import utilities as util  # noqa: E402


class DuplicateStdoutFixTest(unittest.TestCase):
    """Test that duplicate STDOUT messages have been fixed."""

    def setUp(self):
        """Set up test environment."""
        # Clean up data folder to test directory creation
        if os.path.exists('./data'):
            shutil.rmtree('./data')

        # Reset flask server mode
        util.log_stdout_to_stderr = False

    def tearDown(self):
        """Clean up after tests."""
        # Reset flask server mode
        util.log_stdout_to_stderr = False

    def capture_stdout_stderr(self, func):
        """Capture both stdout and stderr output."""
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        stdout_capture = StringIO()
        stderr_capture = StringIO()

        try:
            sys.stdout = stdout_capture
            sys.stderr = stderr_capture
            func()
            return stdout_capture.getvalue(), stderr_capture.getvalue()
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr

    def test_data_log_directory_creation_not_in_stdout(self):
        """Test that DATA_LOG directory creation message doesn't go to STDOUT."""
        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.DATA_LOG)
        )

        # Message should be in STDOUT but directory creation should be in STDERR
        self.assertIn('Test message', stdout)
        self.assertNotIn('data folder', stdout)
        self.assertIn('data folder', stderr)

    def test_dual_stream_log_directory_creation_not_in_stdout(self):
        """Test that DUAL_STREAM_LOG directory creation message doesn't go to STDOUT."""  # noqa: E501
        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.DUAL_STREAM_LOG)
        )

        # Message should be in STDOUT but directory creation should be in STDERR
        self.assertIn('Test message', stdout)
        self.assertNotIn('data folder', stdout)
        self.assertIn('data folder', stderr)

    def test_flask_server_mode_no_duplicate_with_stderr_log(self):
        """Test that flask server mode doesn't duplicate when STDERR_LOG is explicit."""  # noqa: E501
        util.log_stdout_to_stderr = True

        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.STDOUT_LOG | util.STDERR_LOG)
        )

        # Should only appear in STDERR once, not in STDOUT
        self.assertEqual('', stdout)
        self.assertEqual('Test message\n', stderr)

    def test_explicit_stdout_stderr_combination_still_works(self):
        """Test that explicit STDOUT_LOG | STDERR_LOG combination still works when not in flask mode."""  # noqa: E501
        util.log_stdout_to_stderr = False

        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.STDOUT_LOG | util.STDERR_LOG)
        )

        # Should appear in both streams when explicitly requested
        self.assertEqual('Test message\n', stdout)
        self.assertEqual('Test message\n', stderr)

    def test_flask_server_mode_stdout_to_stderr_conversion(self):
        """Test that flask server mode correctly converts STDOUT to STDERR."""
        util.log_stdout_to_stderr = True

        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.STDOUT_LOG)
        )

        # Should only appear in STDERR, not STDOUT
        self.assertEqual('', stdout)
        self.assertEqual('Test message\n', stderr)

    def test_both_log_mode_works_correctly(self):
        """Test that BOTH_LOG mode works correctly without duplicates."""
        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.BOTH_LOG)
        )

        # Should appear in STDOUT only (file logging tested separately)
        # Directory creation message should be in STDERR
        self.assertEqual('Test message\n', stdout)
        self.assertIn('data folder', stderr)

    def test_quiet_log_mode_no_duplicates(self):
        """Test that QUIET_LOG mode doesn't produce duplicates."""
        stdout, stderr = self.capture_stdout_stderr(
            lambda: util.log_msg('Test message', util.QUIET_LOG)
        )

        # Should appear in STDOUT only once
        self.assertEqual('Test message\n', stdout)
        self.assertEqual('', stderr)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
Unit tests for the `flask_generic` module in the `thermostatsupervisor` package.
Classes:
    TestFlaskGeneric: Contains unit tests for the `schedule_ipban_block_list_report`
                      and `print_ipban_block_list_with_timestamp` functions.
Methods:
    test_schedule_ipban_block_list_report(MockAPScheduler):
        Tests the `schedule_ipban_block_list_report` function to ensure it schedules
        the IP ban block list report correctly with different debug modes.
    test_print_ipban_block_list_with_timestamp(mock_datetime):
        Tests the `print_ipban_block_list_with_timestamp` function to ensure it prints
        the IP ban block list with the correct timestamp.
"""

# built-in modules
import datetime
import unittest
from unittest.mock import MagicMock, patch
import unittest.mock

# thrird-party modules

# local modules
from thermostatsupervisor.flask_generic import (
    CustomJSONEncoder,
    clear_ipban_block_list,
    initialize_ipban,
    print_ipban_block_list,
    print_ipban_block_list_with_timestamp,
    print_flask_config,
    schedule_ipban_block_list_report,
    set_flask_cookie_config,
)
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class TestFlaskGeneric(utc.UnitTest):
    """
    Test suite for Flask generic functionalities related to IP ban management.
    This test suite includes the following test cases:
    1. `test_print_ipban_block_list_with_timestamp`:
        - Tests the `print_ipban_block_list_with_timestamp` function to ensure it prints
        - Mocks the current datetime to return a specific timestamp and verifies that
          the print function is called with the expected output string containing the
          timestamp and the IP ban block list.
    2. `test_schedule_ipban_block_list_report_debug_mode`:
        - Tests the `schedule_ipban_block_list_report` function to ensure that it
          schedules.
        - Sets `debug_mode` to True and verifies that the print function is called with
          the expected message indicating that the IP ban BlockList report is scheduled
          every 1.0 minutes.
    3. `test_schedule_ipban_block_list_report_normal_mode`:
        - Tests the `schedule_ipban_block_list_report` function in normal mode.
        - Sets `debug_mode` to False and verifies that the print function is called with
          the expected message indicating that the IP ban BlockList report is scheduled
          every 1440.0 minutes.
    """

    def setUp(self):
        """
        Set up the test environment for the Flask application.
        This method is called before each test is executed. It initializes a mock
        IP ban object and sets up its return value for the get_block_list method.
        Attributes:
            mock_ip_ban (MagicMock): A mock object for simulating IP ban functionality.
        """

        super().setUp()
        self.mock_ip_ban = MagicMock()
        self.mock_ip_ban.get_block_list.return_value = {
            "192.168.1.1": {
                "count": 3,
                "timestamp": datetime.datetime(2024, 1, 1, 12, 0, 0),
            }
        }

    @patch("builtins.print")
    def test_print_ipban_block_list_with_timestamp(self, mock_print):
        """
        Test the print_ipban_block_list_with_timestamp function to ensure it prints
        the IP ban block list with the correct timestamp.
        Args:
            mock_print (Mock): Mock object for the print function.
        Arrange:
            - Set the expected timestamp to "2024-01-01 12:00:00".
            - Mock the current datetime to return the expected timestamp.
        Act:
            - Call the print_ipban_block_list_with_timestamp function with the mocked
              IP ban list.
        Assert:
            - Verify that the print function is called once with the expected output
              string containing the timestamp and the IP ban block list.
        """

        # Arrange
        expected_timestamp = "2024-01-01 12:00:00"
        mock_now = datetime.datetime(2024, 1, 1, 12, 0, 0)

        with patch("datetime.datetime") as mock_datetime:
            mock_datetime.now.return_value = mock_now

            # Act
            print_ipban_block_list_with_timestamp(self.mock_ip_ban)

            # Assert
            mock_print.assert_called_once_with(
                f"{expected_timestamp}: ip_ban block list: "
                f"{self.mock_ip_ban.get_block_list()}"
            )

    @patch("builtins.print")
    def test_schedule_ipban_block_list_report_debug_mode(self, mock_print):
        """
        Test the `schedule_ipban_block_list_report` function to ensure that it schedules
        the IP ban block list report correctly when debug mode is enabled.
        Args:
            mock_print (Mock): Mock object for the print function.
        Arrange:
            - Set `debug_mode` to True.
            - Define the expected interval as "1.0" minutes.
        Act:
            - Patch the `BackgroundScheduler` from `apscheduler.schedulers.background`.
            - Call `schedule_ipban_block_list_report` with the mock IP ban object and
              debug mode.
        Assert:
            - Verify that the print function is called once with the expected message
              indicating that the IP ban BlockList report is scheduled every
              1.0 minutes.
        """

        # Arrange
        debug_mode = True
        expected_interval = "1.0"

        # Act
        with patch("apscheduler.schedulers.background.BackgroundScheduler"):
            schedule_ipban_block_list_report(self.mock_ip_ban, debug_mode)

        # Assert
        mock_print.assert_called_once_with(
            f"ip_ban BlockList report scheduled every {expected_interval} minutes"
        )

    @patch("builtins.print")
    def test_schedule_ipban_block_list_report_normal_mode(self, mock_print):
        """
        Test the `schedule_ipban_block_list_report` function in normal mode.
        This test verifies that the `schedule_ipban_block_list_report` function
        schedules the IP ban block list report correctly when the debug mode is
        set to False. It mocks the `BackgroundScheduler` and checks that the
        `mock_print` function is called with the expected message.
        Args:
            mock_print (Mock): Mock object for the print function.
        Assertions:
            mock_print.assert_called_once_with: Verifies that the print function
            is called once with the expected message indicating the IP ban
            BlockList report is scheduled every 1440.0 minutes.
        """

        # Arrange
        debug_mode = False
        expected_interval = "1440.0"

        # Act
        with patch("apscheduler.schedulers.background.BackgroundScheduler"):
            schedule_ipban_block_list_report(self.mock_ip_ban, debug_mode)

        # Assert
        mock_print.assert_called_once_with(
            f"ip_ban BlockList report scheduled every {expected_interval} minutes"
        )

    def test_custom_json_encoder_default_datetime(self):
        """Test CustomJSONEncoder.default() with datetime object."""
        encoder = CustomJSONEncoder()
        test_datetime = datetime.datetime(2024, 1, 1, 12, 0, 0)

        with patch("builtins.print") as mock_print:
            result = encoder.default(test_datetime)
            self.assertEqual(result, "2024-01-01T12:00:00")
            mock_print.assert_called_once_with(
                f"CustomJSONEncoder enabled: {test_datetime}"
            )

    def test_custom_json_encoder_default_non_datetime(self):
        """Test CustomJSONEncoder.default() with non-datetime object."""
        encoder = CustomJSONEncoder()
        test_obj = {"key": "value"}

        with patch("builtins.print") as mock_print:
            with self.assertRaises(TypeError):
                encoder.default(test_obj)
            mock_print.assert_called_once_with(
                f"CustomJSONEncoder bypassed: {test_obj}, {type(test_obj)}"
            )

    @patch("builtins.print")
    def test_clear_ipban_block_list_all_ips(self, mock_print):
        """Test clear_ipban_block_list() clearing all IP addresses."""
        # Setup mock with multiple IPs
        self.mock_ip_ban.get_block_list.return_value = {
            "192.168.1.1": {"count": 3},
            "192.168.1.2": {"count": 5},
        }
        self.mock_ip_ban.remove.return_value = True

        # Call function with no specific IP address
        clear_ipban_block_list(self.mock_ip_ban)

        # Verify prints and removes were called
        # Should have: before, clearing all, after = 3 prints
        self.assertEqual(mock_print.call_count, 3)
        self.assertEqual(self.mock_ip_ban.remove.call_count, 2)

    @patch("builtins.print")
    def test_clear_ipban_block_list_specific_ip(self, mock_print):
        """Test clear_ipban_block_list() clearing specific IP address."""
        self.mock_ip_ban.remove.return_value = True
        test_ip = "192.168.1.100"

        # Call function with specific IP address
        clear_ipban_block_list(self.mock_ip_ban, test_ip)

        # Verify remove was called with specific IP
        self.mock_ip_ban.remove.assert_called_once_with(test_ip)
        self.assertEqual(mock_print.call_count, 3)  # before, clearing, after

    @patch("builtins.print")
    def test_clear_ipban_block_list_ip_not_found(self, mock_print):
        """Test clear_ipban_block_list() when IP is not found."""
        self.mock_ip_ban.remove.return_value = False
        test_ip = "192.168.1.100"

        # Call function with IP that doesn't exist
        clear_ipban_block_list(self.mock_ip_ban, test_ip)

        # Verify warning message was printed
        warning_calls = [
            call for call in mock_print.call_args_list if "WARNING" in str(call)
        ]
        self.assertTrue(len(warning_calls) > 0)

    @patch("builtins.print")
    def test_print_ipban_block_list(self, mock_print):
        """Test print_ipban_block_list() function."""
        expected_block_list = {"192.168.1.1": {"count": 3}}
        self.mock_ip_ban.get_block_list.return_value = expected_block_list

        print_ipban_block_list(self.mock_ip_ban)

        mock_print.assert_called_once_with(f"ip_ban block list: {expected_block_list}")

    @patch("thermostatsupervisor.flask_generic.IpBan")
    @patch("builtins.print")
    def test_initialize_ipban(self, mock_print, mock_ipban_class):
        """Test initialize_ipban() function."""
        # Setup mock app and ipban instance
        mock_app = MagicMock()
        mock_ipban_instance = MagicMock()
        mock_ipban_class.return_value = mock_ipban_instance

        # Call function
        result = initialize_ipban(mock_app)

        # Verify IpBan was created with correct parameters
        mock_ipban_class.assert_called_once()
        mock_ipban_instance.init_app.assert_called_once_with(mock_app)
        mock_ipban_instance.load_nuisances.assert_called_once()
        self.assertEqual(result, mock_ipban_instance)

    def test_set_flask_cookie_config(self):
        """Test set_flask_cookie_config() function."""
        mock_app = MagicMock()

        set_flask_cookie_config(mock_app)

        mock_app.config.update.assert_called_once_with(
            SESSION_COOKIE_SECURE=True,
            SESSION_COOKIE_HTTPONLY=True,
            SESSION_COOKIE_SAMESITE="Lax",
        )

    @patch("builtins.print")
    def test_print_flask_config(self, mock_print):
        """Test print_flask_config() function."""
        mock_app = MagicMock()
        test_config = {"DEBUG": False, "TESTING": True}
        mock_app.config = test_config

        print_flask_config(mock_app)

        # Verify both print statements were called
        expected_calls = [
            unittest.mock.call("flask config:"),
            unittest.mock.call(f"{test_config}"),
        ]
        mock_print.assert_has_calls(expected_calls)


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""
Integration test for nest environment variable automation functionality.

This test validates that the full initialization process works correctly
with environment variable token cache automation.
"""
import json
import os
import tempfile
import unittest
from unittest.mock import patch

# local imports
from thermostatsupervisor import nest
from tests import unit_test_common as utc


class TestNestEnvIntegration(utc.UnitTest):
    """Integration test for nest environment variable automation."""

    def setUp(self):
        """Set up test fixtures."""
        super().setUp()
        self.print_test_name()

        # Create a temporary directory for cache files
        self.temp_dir = tempfile.mkdtemp()
        self.cache_file_path = os.path.join(self.temp_dir, "test_token_cache.json")

        # Mock required env vars
        self.mock_env_vars = {
            "GCLOUD_CLIENT_ID": "test_client_id",
            "GCLOUD_CLIENT_SECRET": "test_client_secret",
            "DAC_PROJECT_ID": "test_project_id",
            "NEST_ACCESS_TOKEN": "ya29.test_access_token",
            "NEST_REFRESH_TOKEN": "1//040test_refresh_token",
            "NEST_TOKEN_EXPIRES_IN": "3600",
        }

    def tearDown(self):
        """Clean up test fixtures."""
        # Clean up any files created
        if os.path.exists(self.cache_file_path):
            os.unlink(self.cache_file_path)
        os.rmdir(self.temp_dir)

    @patch.dict(os.environ, {}, clear=False)
    def test_integration_token_cache_creation(self):
        """Test integration of token cache creation with environment variables."""
        # Set environment variables
        for key, value in self.mock_env_vars.items():
            os.environ[key] = value

        # Verify cache file doesn't exist initially
        self.assertFalse(os.path.exists(self.cache_file_path))

        # Create minimal thermostat instance and call the method
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.verbose = True

        # Call the token cache creation method
        thermostat._create_token_cache_from_env_if_needed()

        # Verify token cache file was created
        self.assertTrue(os.path.exists(self.cache_file_path))

        # Verify token cache file contents
        with open(self.cache_file_path, "r", encoding="utf-8") as f:
            token_data = json.load(f)

        self.assertEqual(
            token_data["access_token"], self.mock_env_vars["NEST_ACCESS_TOKEN"]
        )
        self.assertEqual(
            token_data["refresh_token"], self.mock_env_vars["NEST_REFRESH_TOKEN"]
        )
        self.assertEqual(
            token_data["expires_in"], int(self.mock_env_vars["NEST_TOKEN_EXPIRES_IN"])
        )

    @patch.dict(os.environ, {}, clear=False)
    def test_no_env_vars_no_cache_created(self):
        """Test no cache file created when environment variables missing."""
        # Set only required OAuth credentials but no token env vars
        os.environ["GCLOUD_CLIENT_ID"] = "test_client_id"
        os.environ["GCLOUD_CLIENT_SECRET"] = "test_client_secret"
        os.environ["DAC_PROJECT_ID"] = "test_project_id"

        # Create minimal thermostat instance
        thermostat = nest.ThermostatClass.__new__(nest.ThermostatClass)
        thermostat.access_token_cache_file = self.cache_file_path
        thermostat.verbose = True

        # Call the token cache creation method
        thermostat._create_token_cache_from_env_if_needed()

        # Verify no cache file was created
        self.assertFalse(os.path.exists(self.cache_file_path))


if __name__ == "__main__":
    unittest.main()



"""
Thermostat API.

This file should be updated for any new thermostats supported and
any changes to thermostat configs.
"""

# built ins
import munch

# local imports
from thermostatsupervisor import blink_config
from thermostatsupervisor import emulator_config
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import kumocloud_config
from thermostatsupervisor import kumocloudv3_config
from thermostatsupervisor import kumolocal_config
from thermostatsupervisor import mmm_config
from thermostatsupervisor import nest_config
from thermostatsupervisor import sht31_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# thermostat types
DEFAULT_THERMOSTAT = emulator_config.ALIAS
DEFAULT_ZONE_NAME = util.default_parent_key

KEY_MISSING_SUFFIX = "_KEY_MISSING>"

# list of thermostat config modules supported
config_modules = [
    blink_config,
    emulator_config,
    honeywell_config,
    kumocloud_config,
    kumocloudv3_config,
    kumolocal_config,
    mmm_config,
    nest_config,
    sht31_config,
]

SUPPORTED_THERMOSTATS = {
    # "module" = module to import
    # "type" = thermostat type index number
    # "zones" = zone numbers supported
    # "modes" = modes supported
}
for config_module in config_modules:
    SUPPORTED_THERMOSTATS.update({config_module.ALIAS: config_module.supported_configs})

# dictionary of required env variables for each thermostat type
thermostats = {}
for config_module in config_modules:
    # Use required_env_variables if explicitly defined, otherwise use env_variables
    # This allows consolidation of duplicate env dicts while preserving special cases
    required_env_vars = getattr(config_module, "required_env_variables", None)
    if required_env_vars is None:
        required_env_vars = getattr(config_module, "env_variables", {})

    thermostats.update(
        {config_module.ALIAS: {"required_env_variables": required_env_vars}}
    )


# runtime override parameters
# note script name is omitted, starting with first parameter
# index 0 (script name) is not included in this dict because it is
# not a runtime argument
input_flds = munch.Munch()
input_flds.thermostat_type = "thermostat_type"
input_flds.zone = "zone"
input_flds.poll_time = "poll_time"
input_flds.connection_time = "connection_time"
input_flds.tolerance = "tolerance"
input_flds.target_mode = "target_mode"
input_flds.measurements = "measurements"
input_flds.input_file = "input_file"

uip = None  # user inputs object


class UserInputs(util.UserInputs):
    """Manage runtime arguments for thermostat_api."""

    def __init__(
        self,
        argv_list=None,
        help_description=None,
        suppress_warnings=False,
        thermostat_type=DEFAULT_THERMOSTAT,
        zone_name=DEFAULT_ZONE_NAME,
    ):
        """Initialize UserInputs for thermostat_api.

        Args:
            argv_list (list, optional): Override runtime values. Defaults to None.
            help_description (str, optional): Description field for help text.
                Defaults to None.
            suppress_warnings (bool, optional): True to suppress warning messages.
                Defaults to False.
            thermostat_type (str, optional): Thermostat type alias.
                Defaults to DEFAULT_THERMOSTAT.
            zone_name (str, optional): Thermostat zone name (e.g. 'living room').
                Defaults to DEFAULT_ZONE_NAME.
        """
        self.argv_list = argv_list
        self.help_description = help_description
        self.suppress_warnings = suppress_warnings
        self.thermostat_type = thermostat_type  # default if not provided
        self.zone_name = zone_name

        # initialize parent class
        super().__init__(argv_list, help_description, suppress_warnings, zone_name)

    def initialize_user_inputs(self, parent_keys=None):
        """Populate user_inputs dictionary with thermostat-specific parameters.

        Args:
            parent_keys (list, optional): List of parent keys to initialize.
                Defaults to [self.default_parent_key].
        """
        if parent_keys is None:
            parent_keys = [self.default_parent_key]
        self.valid_sflags = []
        self.user_inputs = {}  # init
        # define the user_inputs dict.
        for parent_key in parent_keys:
            self.user_inputs[parent_key] = {
                input_flds.thermostat_type: {
                    "order": 1,  # index in the argv list
                    "value": None,
                    "type": str,
                    "default": self.thermostat_type,
                    "valid_range": list(SUPPORTED_THERMOSTATS.keys()),
                    "sflag": "-t",
                    "lflag": "--" + input_flds.thermostat_type,
                    "help": "thermostat type",
                    "required": False,  # default value is set if missing.
                },
                input_flds.zone: {
                    "order": 2,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 0,
                    "valid_range": None,  # updated once thermostat is known
                    "sflag": "-z",
                    "lflag": "--" + input_flds.zone,
                    "help": "target zone number",
                    "required": False,  # defaults to idx 0 in supported zones
                },
                input_flds.poll_time: {
                    "order": 3,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 60 * 10,
                    "valid_range": range(0, 24 * 60 * 60),
                    "sflag": "-p",
                    "lflag": "--" + input_flds.poll_time,
                    "help": "poll time (sec)",
                    "required": False,
                },
                input_flds.connection_time: {
                    "order": 4,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 24 * 60 * 60,  # 1 day
                    "valid_range": range(0, 7 * 24 * 60 * 60),  # up to 1 wk
                    "sflag": "-c",
                    "lflag": "--" + input_flds.connection_time,
                    "help": "server connection time (sec)",
                    "required": False,
                },
                input_flds.tolerance: {
                    "order": 5,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 2,
                    "valid_range": range(0, 10),
                    "sflag": "-d",
                    "lflag": "--" + input_flds.tolerance,
                    "help": "tolerance (°F)",
                    "required": False,
                },
                input_flds.target_mode: {
                    "order": 6,  # index in the argv list
                    "value": None,
                    "type": str,
                    "default": "UNKNOWN_MODE",
                    "valid_range": None,  # updated once thermostat is known
                    "sflag": "-m",
                    "lflag": "--" + input_flds.target_mode,
                    "help": "target thermostat mode",
                    "required": False,
                },
                input_flds.measurements: {
                    "order": 7,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 52560,  # 1 year at 10min polling (365d × 24h × 6)
                    "valid_range": range(1, 52562),  # up to 1 year + 1 measurement
                    "sflag": "-n",
                    "lflag": "--" + input_flds.measurements,
                    "help": "number of measurements",
                    "required": False,
                },
                input_flds.input_file: {
                    "order": 8,  # index in the argv list
                    "value": None,
                    # "type": lambda x: self.is_valid_file(x),
                    "type": str,  # argparse.FileType('r', encoding='UTF-8'),
                    "default": None,
                    "valid_range": None,
                    "sflag": "-f",
                    "lflag": "--" + input_flds.input_file,
                    "help": "input file",
                    "required": False,
                },
            }
            self.valid_sflags += [
                self.user_inputs[parent_key][k]["sflag"]
                for k in self.user_inputs[parent_key].keys()
            ]

    def dynamic_update_user_inputs(self):
        """
        Update thermostat-specific values in user_inputs dict.

        This function expands each input parameter list to match
        the length of the thermostat parameter field.
        """
        self._initialize_parent_keys()
        input_file = self._get_input_file()

        if input_file is not None:
            self._process_input_file(input_file)
        elif self._has_populated_user_inputs():
            self._process_argv_inputs()
        else:
            self._handle_unpopulated_inputs()

        self._finalize_thermostat_configuration()

    def _initialize_parent_keys(self):
        """Initialize section list to single item list of one thermostat."""
        self.parent_keys = [self.default_parent_key]

    def _get_input_file(self):
        """Get input file from user inputs if specified."""
        return self.get_user_inputs(self.default_parent_key, input_flds.input_file)

    def _process_input_file(self, input_file):
        """Process input file and populate user inputs from file."""
        self.using_input_file = True
        self.parse_input_file(input_file)
        self.parent_keys = list(self.user_inputs_file.keys())
        self.initialize_user_inputs(self.parent_keys)
        self._populate_from_file()
        self.zone_name = self.parent_keys[0]

    def _populate_from_file(self):
        """Populate user_inputs from user_inputs_file."""
        for section in self.parent_keys:
            for fld in input_flds:
                if fld == input_flds.input_file:
                    continue  # input file field will not be in the file

                if self.user_inputs[section][fld]["type"] in [int, float, str]:
                    self._cast_field_value(section, fld)
                else:
                    self._read_raw_field_value(section, fld)

    def _cast_field_value(self, section, fld):
        """Cast data type when reading value from file."""
        try:
            cast_value = self.user_inputs[section][fld]["type"](
                self.user_inputs_file[section].get(input_flds[fld])
            )
            self.user_inputs[section][fld]["value"] = cast_value
            # cast original input value in user_inputs_file as well
            self.user_inputs_file[section][input_flds[fld]] = cast_value
        except Exception:
            print(f"exception in section={section}, fld={fld}")
            raise

    def _read_raw_field_value(self, section, fld):
        """Read raw value without casting from file."""
        self.user_inputs[section][fld]["value"] = self.user_inputs_file[section].get(
            input_flds[fld]
        )

    def _has_populated_user_inputs(self):
        """Check if user_inputs has already been populated."""
        return (
            self.get_user_inputs(
                list(self.user_inputs.keys())[0], input_flds.thermostat_type
            )
            is not None
        )

    def _process_argv_inputs(self):
        """Process argv inputs (only currently supporting 1 zone)."""
        current_keys = list(self.user_inputs.keys())
        if len(current_keys) != 1:
            raise KeyError(f"user_input keys={current_keys}, expected only 1 key")

        current_key = current_keys[0]
        new_key = self._build_zone_key(current_key)
        self.user_inputs[new_key] = self.user_inputs.pop(current_key)
        self._update_zone_parameters(new_key)

    def _build_zone_key(self, current_key):
        """Build new zone key as <zone_name>_<zone_number>."""
        thermostat_type = self.get_user_inputs(current_key, input_flds.thermostat_type)
        zone_number = self.get_user_inputs(current_key, input_flds.zone)
        return f"{thermostat_type}_{zone_number}"

    def _update_zone_parameters(self, new_key):
        """Update parameters for new parent keys."""
        self.zone_name = new_key
        self.default_parent_key = new_key
        self.parent_keys = list(self.user_inputs.keys())

    def _handle_unpopulated_inputs(self):
        """Handle case where inputs haven't been populated yet."""
        runtime_args = self.get_user_inputs(
            list(self.user_inputs.keys())[0], input_flds.thermostat_type
        )
        print(f"runtime args: {runtime_args}")

    def _finalize_thermostat_configuration(self):
        """Set up thermostat configuration for all zones."""
        for zone_name in self.parent_keys:
            thermostat_type = self._get_thermostat_type(zone_name)
            self._configure_zone_ranges(zone_name, thermostat_type)

    def _get_thermostat_type(self, zone_name):
        """Get thermostat type for zone, defaulting if not set."""
        thermostat_type = self.get_user_inputs(zone_name, input_flds.thermostat_type)
        return thermostat_type if thermostat_type is not None else self.thermostat_type

    def _configure_zone_ranges(self, zone_name, thermostat_type):
        """Configure valid ranges for zone and target mode fields."""
        self._set_zone_valid_range(zone_name, thermostat_type)
        self._set_target_mode_valid_range(zone_name, thermostat_type)

    def _set_zone_valid_range(self, zone_name, thermostat_type):
        """Set valid range for zone field."""
        try:
            self.user_inputs[zone_name][input_flds.zone][
                "valid_range"
            ] = SUPPORTED_THERMOSTATS[thermostat_type]["zones"]
        except KeyError:
            print(
                f"\nKeyError: one or more keys are invalid (zone_name="
                f"{zone_name}, zone_number={input_flds.zone}, "
                f"thermostat_type={thermostat_type})\n"
            )
            raise

    def _set_target_mode_valid_range(self, zone_name, thermostat_type):
        """Set valid range for target mode field."""
        try:
            self.user_inputs[zone_name][input_flds.target_mode][
                "valid_range"
            ] = SUPPORTED_THERMOSTATS[thermostat_type]["modes"]
        except KeyError:
            print(
                f"\nKeyError: one or more keys are invalid (zone_name="
                f"{zone_name}, target_mode={input_flds.target_mode}, "
                f"thermostat_type={thermostat_type})\n"
            )
            raise

    def max_measurement_count_exceeded(self, measurement):
        """
        Return True if max measurement reached.

        inputs:
            measurement(int): current measurement value
        returns:
            (bool): True if max measurement reached.
        """
        max_measurements = self.get_user_inputs(self.zone_name, "measurements")
        if max_measurements is None:
            return False
        elif measurement > max_measurements:
            return True
        else:
            return False


def verify_required_env_variables(tstat, zone_str, verbose=True):
    """Verify all required environment variables are present for thermostat config.

    Checks that all required environment variables for the specified thermostat type
    are present in the environment. Some environment variables may require
    zone-specific suffixes (indicated by trailing underscore).

    Args:
        tstat (str): Thermostat type alias (e.g., 'honeywell', 'emulator')
        zone_str (str): Zone identifier as string
        verbose (bool, optional): Enable debug output. Defaults to True.

    Returns:
        bool: True if all required environment variables are present

    Raises:
        KeyError: If required environment variables are missing
    """
    if verbose:
        print("\nchecking required environment variables:")
    key_status = True  # default, all keys present
    for key in thermostats[tstat]["required_env_variables"]:
        # any env key ending in '_' should have zone number appended to it.
        if key[-1] == "_":
            # append zone info to key
            key = key + str(zone_str)
        if verbose:
            print(f"checking required environment key: {key}...", end="")
        env.env_variables[key] = env.get_env_variable(key)["value"]
        if env.env_variables[key] is not None:
            if verbose:
                print("OK")
        else:
            util.log_msg(
                f"{tstat}: zone {zone_str}: FATAL error: one or more required"
                f" environemental keys are missing, exiting program",
                mode=util.BOTH_LOG,
            )
            key_status = False
            raise KeyError
    if verbose:
        print("\n")
    return key_status


def load_hardware_library(thermostat_type):
    """Dynamically load appropriate hardware library for the thermostat type.

    Imports the hardware-specific module for the given thermostat type using dynamic
    module loading. This allows the system to support multiple thermostat brands
    without static imports.

    Args:
        thermostat_type (str): Thermostat type alias (e.g., 'honeywell',
            'emulator')

    Returns:
        module: Loaded Python module for the thermostat type

    Raises:
        ImportError: If the thermostat type module cannot be loaded
        KeyError: If the thermostat type is not supported
    """
    pkg_name = (
        util.PACKAGE_NAME + "." + SUPPORTED_THERMOSTATS[thermostat_type]["module"]
    )
    mod = env.dynamic_module_import(pkg_name)
    return mod


def load_user_inputs(config_mod):
    """
    Load the default user inputs and return the zone number.

    inputs:
        config_mod(obj): config module
    returns:
        zone_number(int): zone number
    """
    global uip
    zone_name = config_mod.default_zone_name
    uip = UserInputs(
        argv_list=config_mod.argv, thermostat_type=config_mod.ALIAS, zone_name=zone_name
    )
    zone_number = uip.get_user_inputs(uip.zone_name, input_flds.zone)
    return zone_number



"""
Integration test module for kumocloud.py.

This test requires connection to Kumocloud thermostat.
"""

# built-in imports
import unittest

# local imports
# conditionally import kumocloud module to handle missing pykumo dependency
try:
    from thermostatsupervisor import kumocloud
    from thermostatsupervisor import kumocloud_config

    kumocloud_import_error = None
except ImportError as ex:
    # pykumo library not available, tests will be skipped
    kumocloud = None
    kumocloud_config = None
    kumocloud_import_error = ex

from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in kumocloud.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "kumocloud",  # thermostat
            "0",  # zone
            "30",  # poll time in sec
            "1000",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "3",  # number of measurements
        ]
        self.mod = kumocloud
        self.mod_config = kumocloud_config


@unittest.skipIf(
    not utc.ENABLE_KUMOCLOUD_TESTS or kumocloud_import_error,
    "kumocloud tests are disabled",
)
class FunctionalIntegrationTest(IntegrationTest, utc.FunctionalIntegrationTest):
    """
    Test functional performance of kumocloud.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.trait_field = None
        self.metadata_field = "address"
        self.metadata_type = str


@unittest.skipIf(
    not utc.ENABLE_KUMOCLOUD_TESTS or kumocloud_import_error,
    "kumocloud tests are disabled",
)
class SuperviseIntegrationTest(IntegrationTest, utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of kumocloud.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


@unittest.skipIf(
    not utc.ENABLE_KUMOCLOUD_TESTS or kumocloud_import_error,
    "kumocloud tests are disabled",
)
class PerformanceIntegrationTest(IntegrationTest, utc.PerformanceIntegrationTest):
    """
    Test performance of kumocloud.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = 30
        self.timing_measurements = 30

        # temperature and humidity repeatability measurements
        # temperature and humidity data are int values
        # settings below are tuned for 12 minutes, 4 measurements per minute.
        self.temp_stdev_limit = 2.0  # 1 sigma temp repeatability limit in F
        self.temp_repeatability_measurements = 48  # number of temp msmts.
        self.humidity_stdev_limit = 2.0  # 1 sigma humid repeat. limit %RH
        self.humidity_repeatability_measurements = 48  # number of temp msmts.
        self.poll_interval_sec = 15  # delay between repeatability measurements


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Unit test module for enhanced 403 error handling in sht31.py.

These tests verify that the enhanced 403 error messages provide better
feedback when IP addresses are blocked by ipban.
"""
# built-in imports
import unittest
from unittest.mock import patch, MagicMock
import sys

# Mock the RPi.GPIO import before importing sht31
# This needs to be done before importing sht31 module
sys.modules["RPi"] = MagicMock()
sys.modules["RPi.GPIO"] = MagicMock()
sys.modules["smbus2"] = MagicMock()

# local imports (must come after mocking)
from thermostatsupervisor import sht31  # noqa: E402
from thermostatsupervisor import sht31_config  # noqa: E402
from tests import unit_test_common as utc  # noqa: E402


@unittest.skipIf(False, "Test always enabled for 403 error validation")
class Enhanced403ErrorTest(utc.UnitTest):
    """Test enhanced 403 error handling in sht31.py."""

    def setUp(self):
        """Set up test fixtures."""
        self.test_zone = sht31_config.UNIT_TEST_ZONE
        # Use unit test path to avoid real server connections
        self.thermostat = sht31.ThermostatClass(
            zone=self.test_zone, path=sht31_config.flask_folder.unit_test, verbose=False
        )
        self.thermostat_zone = sht31.ThermostatZone(self.thermostat, verbose=False)

    @patch("thermostatsupervisor.sht31.requests.get")
    def test_ipban_403_error_thermostat_class(self, mock_get):
        """Test enhanced 403 error message for ipban block in ThermostatClass."""
        # Mock a 403 response with ipban message
        mock_response = MagicMock()
        mock_response.text = (
            "403 Forbidden: You don't have the permission to access the "
            "requested resource. It is either read-protected or not "
            "readable by the server."
        )
        mock_get.return_value = mock_response

        # Test that the enhanced error message is generated
        with self.assertRaises(RuntimeError) as context:
            self.thermostat.get_metadata(zone=self.test_zone, retry=False)

        error_message = str(context.exception)
        # Verify the enhanced error message contains key information
        self.assertIn("IP address", error_message)
        self.assertIn("blocked due to suspicious activity", error_message)
        self.assertIn("IP ban protection mechanism", error_message)
        self.assertIn("clear_block_list endpoint", error_message)

    @patch("thermostatsupervisor.sht31.requests.get")
    def test_generic_403_error_thermostat_class(self, mock_get):
        """Test generic 403 error message for non-ipban 403 errors."""
        # Mock a 403 response without ipban message
        mock_response = MagicMock()
        mock_response.text = "403 Forbidden: Access denied for other reasons"
        mock_get.return_value = mock_response

        # Test that the generic error message is generated
        with self.assertRaises(RuntimeError) as context:
            self.thermostat.get_metadata(zone=self.test_zone, retry=False)

        error_message = str(context.exception)
        # Verify the generic error message
        self.assertIn("FATAL ERROR 403: client is forbidden", error_message)
        self.assertNotIn("IP ban protection mechanism", error_message)

    @patch("thermostatsupervisor.sht31.requests.get")
    def test_ipban_403_error_thermostat_zone(self, mock_get):
        """Test enhanced 403 error message for ipban block in ThermostatZone."""
        # Mock a 403 response with ipban message
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.text = (
            "You don't have the permission to access the requested resource. "
            "It is either read-protected or not readable by the server."
        )
        mock_get.return_value = mock_response

        # Test that the enhanced error message is generated
        with self.assertRaises(RuntimeError) as context:
            self.thermostat_zone.get_metadata(retry=False)

        error_message = str(context.exception)
        # Verify the enhanced error message contains key information
        self.assertIn("IP address", error_message)
        self.assertIn("blocked due to suspicious activity", error_message)
        self.assertIn("IP ban protection mechanism", error_message)
        self.assertIn("clear_block_list endpoint", error_message)

    @patch("thermostatsupervisor.sht31.requests.get")
    def test_generic_403_error_thermostat_zone(self, mock_get):
        """Test generic 403 error message for non-ipban 403 errors."""
        # Mock a 403 response without ipban message
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.text = "Access denied for other reasons"
        mock_get.return_value = mock_response

        # Test that the generic error message is generated
        with self.assertRaises(RuntimeError) as context:
            self.thermostat_zone.get_metadata(retry=False)

        error_message = str(context.exception)
        # Verify the generic error message
        self.assertIn("FATAL ERROR 403: client is forbidden", error_message)
        self.assertNotIn("IP ban protection mechanism", error_message)

    @patch("socket.gethostbyname")
    @patch("socket.gethostname")
    @patch("thermostatsupervisor.sht31.requests.get")
    def test_ip_address_detection(self, mock_get, mock_hostname, mock_gethostbyname):
        """Test IP address detection in error message."""
        # Mock socket functions
        mock_hostname.return_value = "test-host"
        mock_gethostbyname.return_value = "192.168.1.100"

        # Mock a 403 response with ipban message
        mock_response = MagicMock()
        mock_response.text = (
            "403 Forbidden: You don't have the permission to access the "
            "requested resource. It is either read-protected or not "
            "readable by the server."
        )
        mock_get.return_value = mock_response

        # Test that the IP address is included in the error message
        with self.assertRaises(RuntimeError) as context:
            self.thermostat.get_metadata(zone=self.test_zone, retry=False)

        error_message = str(context.exception)
        self.assertIn("192.168.1.100", error_message)

    @patch("socket.gethostbyname", side_effect=OSError("Network error"))
    @patch("socket.gethostname", side_effect=OSError("Network error"))
    @patch("thermostatsupervisor.sht31.requests.get")
    def test_ip_address_fallback(self, mock_get, mock_hostname, mock_gethostbyname):
        """Test fallback when IP address detection fails."""
        # Mock a 403 response with ipban message
        mock_response = MagicMock()
        mock_response.text = (
            "403 Forbidden: You don't have the permission to access the "
            "requested resource. It is either read-protected or not "
            "readable by the server."
        )
        mock_get.return_value = mock_response

        # Test that fallback IP is used when detection fails
        with self.assertRaises(RuntimeError) as context:
            self.thermostat.get_metadata(zone=self.test_zone, retry=False)

        error_message = str(context.exception)
        self.assertIn("IP address unknown", error_message)


if __name__ == "__main__":
    unittest.main()



"""Generic Flask functionality."""

# built-in libraries
import datetime
import json

# third party libraries
from flask_apscheduler import APScheduler
from flask_ipban import IpBan

# local imports

# ipban
ipban_ban_count = 1
ipban_ban_seconds = 3600 * 24 * 7  # 1 wk
ipban_persistent = False  # True to persist across restarts


# custom JSON encoder to convert datetime objects to isoformat
class CustomJSONEncoder(json.JSONEncoder):
    """
    A custom JSON encoder that handles datetime objects.
    This encoder extends the default JSONEncoder to serialize datetime objects
    into ISO 8601 formatted strings.
    """

    def default(self, o):
        if isinstance(o, datetime.datetime):
            print(f"CustomJSONEncoder enabled: {o}")
            return o.isoformat()
        else:
            print(f"CustomJSONEncoder bypassed: {o}, {type(o)}")
        return super().default(o)


def initialize_ipban(app):
    """
    Initialize ipban agent for blocking hacking attempts.

    inputs:
        app (flask app object)
    returns:
        (ip_ban object)
    """
    # setup ipban
    ip_ban = IpBan(
        app=app,
        ban_count=ipban_ban_count,
        ban_seconds=ipban_ban_seconds,
        persist=ipban_persistent,
    )

    ip_ban.init_app(app)
    ip_ban.load_nuisances()
    print_ipban_block_list(ip_ban)
    return ip_ban


def clear_ipban_block_list(ip_ban, ip_address=None):
    """
    Clear specified IP address from the ip_ban block list.

    inputs:
        ip_ban(ip_ban object)
        ip_address(string): IP address to clear from the block list.
    returns:
        None
    """
    print(f"ip_ban block list before: {ip_ban.get_block_list()}")
    if ip_address is None:
        # clear all ip addresses
        print("clearing all ip addresses from block list")
        for ip in ip_ban.get_block_list():
            if not ip_ban.remove(ip):
                print(f"WARNING: {ip} not found in block list")
    else:
        # clear one ip address
        print(f"clearing ip address {ip_address} from block list")
        if not ip_ban.remove(ip_address):
            print(f"WARNING: {ip_address} not found in block list")
    print(f"ip_ban block list after: {ip_ban.get_block_list()}")


def print_ipban_block_list(ip_ban):
    """
    Print the current ip_ban block list to the console.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    # s = ""
    # s += "<table class='table'><thead>\n"
    # s += ("<tr><th>ip</th><th>count</th><th>permanent</th><th>url</th><th>"
    #       f"timestamp</th></tr>\n")
    # s += ""</thead><tbody>\n"
    # for k, r in ip_ban.get_block_list().items():
    #     s += (f"<tr><td>{k}</td><td>{r['count']}</td><td>"
    #           f"{r.get('permanent', '')}</td><td>{r.get('url', '')}</td><td>"
    #           f"{r['timestamp']}</td></tr>\n")
    # print(f"{s}")
    print(f"ip_ban block list: {ip_ban.get_block_list()}")


def print_ipban_block_list_with_timestamp(ip_ban):
    """
    Print the current ip_ban block list to the console with timestamp.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    now = datetime.datetime.now()
    now_str = now.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{now_str}: ip_ban block list: {ip_ban.get_block_list()}")


def schedule_ipban_block_list_report(ip_ban, debug_mode=False):
    """
    Schedule an ip_ban blocked ip list report.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    # interval = 1 day std, 1 min in debug mode
    interval_sec = 60 * [60 * 24, 1][debug_mode]
    print(f"ip_ban BlockList report scheduled every {interval_sec / 60.0} minutes")
    scheduler = APScheduler()
    kwargs = {"ip_ban": ip_ban}
    scheduler.add_job(
        id="ip_ban BlockList report",
        func=print_ipban_block_list_with_timestamp,
        kwargs=kwargs,
        trigger="interval",
        seconds=interval_sec,
    )
    scheduler.start()


def set_flask_cookie_config(app):
    """
    Set cookie config to protect against cookie attack vectors in
    the Flask configuration.

    inputs:
        app (flask app object)
    returns:
        None
    """
    app.config.update(
        SESSION_COOKIE_SECURE=True,
        SESSION_COOKIE_HTTPONLY=True,
        SESSION_COOKIE_SAMESITE="Lax",
    )


def print_flask_config(app):
    """
    Prints the configuration settings of a Flask application.

    Args:
        app (Flask): The Flask application instance whose configuration is to be
                     printed.
    """
    print("flask config:")
    print(f"{app.config}")



"""
nest config file.
"""

ALIAS = "nest"

# thermostat zones
MAIN_LEVEL = 0  # zone 0
BASEMENT = 1  # zone 1

# constants
MAX_HEAT_SETPOINT = 69.0
MIN_COOL_SETPOINT = 70.0

# Safety temperature settings
# These are used when thermostat is OFF and normal setpoints are unavailable
# Users can adjust these values based on their comfort and safety requirements
SAFETY_HEAT_TEMPERATURE = 45.0  # Minimum safe temperature (°F)
SAFETY_COOL_TEMPERATURE = 75.0  # Maximum safe temperature (°F)

# all environment variables specific to this thermostat type
env_variables = {
    "GCLOUD_CLIENT_ID": None,
    "GCLOUD_CLIENT_SECRET": None,
    "DAC_PROJECT_ID": None,
    "NEST_ACCESS_TOKEN": None,
    "NEST_REFRESH_TOKEN": None,
    "NEST_TOKEN_EXPIRES_IN": None,
}

# min required env variables on all runs
required_env_variables = {}

# supported thermostat configs
supported_configs = {
    "module": "nest",
    "type": 7,
    "zones": [MAIN_LEVEL, BASEMENT],
    "modes": [
        "OFF_MODE",
        "HEAT_MODE",
        "COOL_MODE",
        "DRY_MODE",
        "AUTO_MODE",
        "UNKNOWN_MODE",
        "MANUAL_ECO",
    ],
    "zip_code": "55760",  # Zip code for outdoor weather data
}

# metadata dict
# 'zone_name' is a placeholder, used at Thermostat class level.
# 'zone_name' is updated by device memory via Zone.get_zone_name()
# 'host_name' is used for DNS lookup to determine if device
# 'ip_address' is just for reference.
metadata = {
    MAIN_LEVEL: {
        "ip_address": "192.168.86.229",  # local IP, for ref only.
        "zone_name": "Main Level Thermostat",  # customize your site.
        "host_name": "tbd",  # used for DNS lookup
    },
    BASEMENT: {
        "ip_address": "192.168.86.236",  # local IP, for ref only.
        "zone_name": "Basement Thermostat",  # customize for your site.
        "host_name": "tbd",  # used for DNS lookup
    },
}


def get_available_zones():
    """
    Return list of available zones.

    for this thermostat type, available zones is all zones.

    inputs:
        None.
    returns:
        (list) available zones.
    """
    return supported_configs["zones"]


default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "16",  # poll time in sec
    "356",  # reconnect time in sec
    "4",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
]

# force use of credentials.json file for credentials
use_credentials_file = False
credentials_file_location = ".//credentials.json"

# data caching parameters
# 20 sec. cache period needed to avoid spamming nest server.
cache_period_sec = 20.0  # cache period for data, min 5 sec.
cache_file_location = ".//token_cache.json"  # oauth credentials

# flag to check thermostat response time during basic checkout
check_response_time = False



"""
Integration test module for kumolocal.py.

This test requires connection to kumolocal thermostat.
"""

# built-in imports
import unittest

# local imports
from thermostatsupervisor import kumolocal
from thermostatsupervisor import kumolocal_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in kumolocal.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "kumolocal",  # thermostat
            "0",  # zone
            "20",  # poll time in sec
            "50",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "6",  # number of measurements
        ]
        self.mod = kumolocal
        self.mod_config = kumolocal_config


@unittest.skipIf(not utc.ENABLE_KUMOLOCAL_TESTS, "kumolocal tests are disabled")
class LocalNetworkDetectionTest(IntegrationTest):
    """
    Test local network detection functionality for kumolocal.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()

    def test_local_network_detection_method(self):
        """Test that local network detection method exists and returns bool."""
        # This test verifies the API exists, but doesn't require actual devices
        try:
            thermostat = self.mod.ThermostatClass(zone=0, verbose=False)
            # Test that the method exists and returns a boolean
            result = thermostat.is_local_network_available()
            self.assertIsInstance(result, bool)
        except Exception as e:
            # If kumolocal isn't available, just verify the method signature
            self.skipTest(f"Kumolocal not available for testing: {e}")


@unittest.skipIf(not utc.ENABLE_KUMOLOCAL_TESTS, "kumolocal tests are disabled")
class FunctionalIntegrationTest(IntegrationTest, utc.FunctionalIntegrationTest):
    """
    Test functional performance of kumolocal.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.trait_field = None
        self.metadata_field = "status"
        self.metadata_type = dict


@unittest.skipIf(not utc.ENABLE_KUMOLOCAL_TESTS, "kumolocal tests are disabled")
class SuperviseIntegrationTest(IntegrationTest, utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of kumolocal.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


@unittest.skipIf(not utc.ENABLE_KUMOLOCAL_TESTS, "kumolocal tests are disabled")
class PerformanceIntegrationTest(IntegrationTest, utc.PerformanceIntegrationTest):
    """
    Test performance of kumolocal.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = 30
        self.timing_measurements = 30

        # temperature and humidity repeatability measurements
        # temperature and humidity data are int values
        # settings below are tuned for 12 minutes, 4 measurements per minute.
        self.temp_stdev_limit = 2.0  # 1 sigma temp repeatability limit in F
        self.temp_repeatability_measurements = 48  # number of temp msmts.
        self.humidity_stdev_limit = 2.0  # 1 sigma humid repeat. limit %RH
        self.humidity_repeatability_measurements = 48  # number of temp msmts.
        self.poll_interval_sec = 15  # delay between repeatability measurements


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)
